from typing import Callable, Collection, List, Optional, Tuple

import pytest
from eth_tester.exceptions import TransactionFailed
from eth_typing import HexAddress
from hexbytes import HexBytes
from web3 import Web3
from web3.contract import Contract
from web3.types import Nonce

from raiden_contracts.constants import TEST_SETTLE_TIMEOUT_MIN, ChannelState, MessageTypeId
from raiden_contracts.tests.utils import (
    EMPTY_ADDITIONAL_HASH,
    EMPTY_BALANCE_HASH,
    EMPTY_SIGNATURE,
    LOCKSROOT_OF_NO_LOCKS,
    NONEXISTENT_LOCKSROOT,
    ChannelValues,
    LockedAmounts,
    call_and_transact,
    fake_bytes,
    get_onchain_settlement_amounts,
    get_participants_hash,
    get_settlement_amounts,
)
from raiden_contracts.tests.utils.blockchain import mine_blocks
from raiden_contracts.tests.utils.constants import DEPLOYER_ADDRESS, OnchainBalanceProof
from raiden_contracts.utils.proofs import (
    hash_balance_data,
    sign_balance_proof,
    sign_balance_proof_message,
    sign_withdraw_message,
)
from raiden_contracts.utils.type_aliases import (
    AdditionalHash,
    BalanceHash,
    BlockExpiration,
    ChannelID,
    Locksroot,
    Signature,
    TokenAmount,
)


@pytest.fixture(scope="session")
def create_channel(token_network: Contract) -> Callable:
    def get(A: HexAddress, B: HexAddress, settle_timeout: int = TEST_SETTLE_TIMEOUT_MIN) -> Tuple:
        # Make sure there is no channel existent on chain
        assert token_network.functions.getChannelIdentifier(A, B).call() == 0

        # Open the channel and retrieve the channel identifier
        txn_hash = call_and_transact(
            token_network.functions.openChannel(
                participant1=A, participant2=B, settle_timeout=settle_timeout
            )
        )

        # Get the channel identifier
        channel_identifier = token_network.functions.getChannelIdentifier(A, B).call()

        # Test the channel state on chain
        (
            channel_settle_timeout,
            channel_state,
        ) = token_network.functions.getChannelInfo(channel_identifier, A, B).call()
        assert channel_settle_timeout == settle_timeout
        assert channel_state == ChannelState.OPENED

        return (channel_identifier, txn_hash)

    return get


@pytest.fixture()
def assign_tokens(token_network: Contract, custom_token: Contract) -> Callable:
    def get(participant: HexAddress, deposit: int) -> None:
        owner = DEPLOYER_ADDRESS
        balance = custom_token.functions.balanceOf(participant).call()
        owner_balance = custom_token.functions.balanceOf(owner).call()
        amount = max(deposit - balance, 0)
        transfer_from_owner = min(amount, owner_balance)

        call_and_transact(
            custom_token.functions.transfer(_to=participant, _value=transfer_from_owner),
            {"from": owner},
        )
        assert custom_token.functions.balanceOf(participant).call() >= transfer_from_owner

        if amount > owner_balance:
            minted = amount - owner_balance
            call_and_transact(custom_token.functions.mint(minted), {"from": participant})
        assert custom_token.functions.balanceOf(participant).call() >= deposit
        call_and_transact(
            custom_token.functions.approve(_spender=token_network.address, _value=deposit),
            {"from": participant},
        )
        assert (
            custom_token.functions.allowance(
                _owner=participant, _spender=token_network.address
            ).call()
            >= deposit
        )

    return get


@pytest.fixture()
def channel_deposit(token_network: Contract, assign_tokens: Callable) -> Callable:
    def get(
        channel_identifier: int,
        participant: HexAddress,
        deposit: int,
        partner: HexAddress,
        tx_from: Optional[HexAddress] = None,
    ) -> HexBytes:
        tx_from = tx_from or participant
        assign_tokens(tx_from, deposit)

        txn_hash = call_and_transact(
            token_network.functions.setTotalDeposit(
                channel_identifier=channel_identifier,
                participant=participant,
                total_deposit=deposit,
                partner=partner,
            ),
            {"from": tx_from},
        )
        return txn_hash

    return get


@pytest.fixture()
def create_channel_and_deposit(create_channel: Callable, channel_deposit: Callable) -> Callable:
    def get(
        participant1: HexAddress,
        participant2: HexAddress,
        deposit1: int = 0,
        deposit2: int = 0,
        settle_timeout: int = TEST_SETTLE_TIMEOUT_MIN,
    ) -> int:
        channel_identifier = create_channel(participant1, participant2, settle_timeout)[0]

        if deposit1 > 0:
            channel_deposit(channel_identifier, participant1, deposit1, participant2)
        if deposit2 > 0:
            channel_deposit(channel_identifier, participant2, deposit2, participant1)
        return channel_identifier

    return get


@pytest.fixture()
def withdraw_channel(token_network: Contract, create_withdraw_signatures: Callable) -> Callable:
    def get(
        channel_identifier: int,
        participant: HexAddress,
        withdraw_amount: int,
        expiration_block: int,
        partner: HexAddress,
        delegate: Optional[HexAddress] = None,
    ) -> HexBytes:
        delegate = delegate or participant
        channel_identifier = token_network.functions.getChannelIdentifier(
            participant, partner
        ).call()

        (signature_participant, signature_partner) = create_withdraw_signatures(
            [participant, partner],
            channel_identifier,
            participant,
            withdraw_amount,
            expiration_block,
        )
        txn_hash = call_and_transact(
            token_network.functions.setTotalWithdraw(
                channel_identifier,
                participant,
                withdraw_amount,
                expiration_block,
                signature_participant,
                signature_partner,
            ),
            {"from": delegate},
        )
        return txn_hash

    return get


@pytest.fixture()
def cooperative_settle_channel(
    token_network: Contract, create_withdraw_signatures: Callable
) -> Callable:
    def get(
        channel_identifier: int,
        participant: HexAddress,
        partner: HexAddress,
        withdraw_amount_participant: int,
        withdraw_amount_partner: int,
        expiration_block: int,
        delegate: Optional[HexAddress] = None,
    ) -> HexBytes:
        delegate = delegate or participant
        channel_identifier = token_network.functions.getChannelIdentifier(
            participant, partner
        ).call()

        (signature_A1, signature_B1) = create_withdraw_signatures(
            [participant, partner],
            channel_identifier,
            participant,
            withdraw_amount_participant,
            expiration_block,
        )
        (signature_A2, signature_B2) = create_withdraw_signatures(
            [participant, partner],
            channel_identifier,
            partner,
            withdraw_amount_partner,
            expiration_block,
        )

        txn_hash = call_and_transact(
            token_network.functions.cooperativeSettle(
                channel_identifier,
                (
                    participant,
                    withdraw_amount_participant,
                    expiration_block,
                    signature_A1,
                    signature_B1,
                ),
                (partner, withdraw_amount_partner, expiration_block, signature_B2, signature_A2),
            ),
            {"from": delegate},
        )
        return txn_hash

    return get


@pytest.fixture()
def close_and_update_channel(
    token_network: Contract,
    create_balance_proof: Callable,
    create_balance_proof_countersignature: Callable,
) -> Callable:
    def get(
        channel_identifier: int,
        participant1: HexAddress,
        participant1_values: ChannelValues,
        participant2: HexAddress,
        participant2_values: ChannelValues,
    ) -> None:
        nonce1 = 5
        nonce2 = 7
        additional_hash1 = fake_bytes(32)
        additional_hash2 = fake_bytes(32)

        balance_proof_1 = create_balance_proof(
            channel_identifier,
            participant1,
            participant1_values.transferred,
            participant1_values.locked_amounts.locked,
            nonce1,
            participant1_values.locksroot,
            additional_hash1,
        )
        balance_proof_2 = create_balance_proof(
            channel_identifier,
            participant2,
            participant2_values.transferred,
            participant2_values.locked_amounts.locked,
            nonce2,
            participant2_values.locksroot,
            additional_hash2,
        )
        balance_proof_close_signature_1 = create_balance_proof_countersignature(
            participant=participant1,
            channel_identifier=channel_identifier,
            msg_type=MessageTypeId.BALANCE_PROOF,
            **balance_proof_2._asdict(),
        )
        balance_proof_update_signature_2 = create_balance_proof_countersignature(
            participant=participant2,
            channel_identifier=channel_identifier,
            msg_type=MessageTypeId.BALANCE_PROOF_UPDATE,
            **balance_proof_1._asdict(),
        )

        call_and_transact(
            token_network.functions.closeChannel(
                channel_identifier,
                participant2,
                participant1,
                *balance_proof_2._asdict().values(),
                balance_proof_close_signature_1,
            ),
            {"from": participant1},
        )

        call_and_transact(
            token_network.functions.updateNonClosingBalanceProof(
                channel_identifier,
                participant1,
                participant2,
                *balance_proof_1._asdict().values(),
                balance_proof_update_signature_2,
            ),
            {"from": participant2},
        )

    return get


@pytest.fixture()
def create_settled_channel(
    web3: Web3,
    token_network: Contract,
    create_channel_and_deposit: Callable,
    close_and_update_channel: Callable,
) -> Callable:
    def get(
        participant1: HexAddress,
        locked_amount1: int,
        locksroot1: bytes,
        participant2: HexAddress,
        locked_amount2: int,
        locksroot2: bytes,
        settle_timeout: int = TEST_SETTLE_TIMEOUT_MIN,
    ) -> int:
        participant1_values = ChannelValues(
            transferred=5,
            locked_amounts=LockedAmounts(claimable_locked=locked_amount1),
            locksroot=locksroot1,
        )
        participant2_values = ChannelValues(
            transferred=40,
            locked_amounts=LockedAmounts(claimable_locked=locked_amount2),
            locksroot=locksroot2,
        )

        participant1_values.deposit = (
            participant1_values.locked_amounts.locked + participant1_values.transferred - 5
        )
        participant2_values.deposit = (
            participant2_values.locked_amounts.locked + participant2_values.transferred + 5
        )

        channel_identifier = create_channel_and_deposit(
            participant1,
            participant2,
            participant1_values.deposit,
            participant2_values.deposit,
            settle_timeout,
        )

        close_and_update_channel(
            channel_identifier,
            participant1,
            participant1_values,
            participant2,
            participant2_values,
        )

        mine_blocks(web3, settle_timeout)

        call_settle(
            token_network,
            channel_identifier,
            participant1,
            participant1_values,
            participant2,
            participant2_values,
        )

        return channel_identifier

    return get


@pytest.fixture()
def reveal_secrets(web3: Web3, secret_registry_contract: Contract) -> Callable:
    def get(tx_from: HexAddress, transfers: List[Tuple]) -> None:
        for (expiration, _, secrethash, secret) in transfers:
            assert web3.eth.block_number < expiration
            call_and_transact(
                secret_registry_contract.functions.registerSecret(secret),
                {"from": tx_from},
            )
            assert (
                secret_registry_contract.functions.getSecretRevealBlockHeight(secrethash).call()
                == web3.eth.block_number
            )

    return get


def common_settle_state_tests(
    custom_token: Contract,
    token_network: Contract,
    channel_identifier: int,
    A: HexAddress,
    balance_A: int,
    B: HexAddress,
    balance_B: int,
    pre_account_balance_A: int,
    pre_account_balance_B: int,
    pre_balance_contract: int,
) -> None:
    # Make sure the correct amount of tokens has been transferred
    account_balance_A = custom_token.functions.balanceOf(A).call()
    account_balance_B = custom_token.functions.balanceOf(B).call()
    balance_contract = custom_token.functions.balanceOf(token_network.address).call()
    assert account_balance_A == pre_account_balance_A + balance_A
    assert account_balance_B == pre_account_balance_B + balance_B
    assert balance_contract == pre_balance_contract - balance_A - balance_B

    # Make sure channel data has been removed
    assert (
        token_network.functions.participants_hash_to_channel_identifier(
            get_participants_hash(A, B)
        ).call()
        == 0
    )

    # Make sure participant data has been removed
    (
        A_deposit,
        A_withdrawn,
        A_is_the_closer,
        A_balance_hash,
        A_nonce,
        _,
        _,
    ) = token_network.functions.getChannelParticipantInfo(channel_identifier, A, B).call()
    assert A_deposit == 0
    assert A_withdrawn == 0
    assert A_is_the_closer == 0
    assert A_balance_hash == fake_bytes(32)
    assert A_nonce == 0

    (
        B_deposit,
        B_withdrawn,
        B_is_the_closer,
        B_balance_hash,
        B_nonce,
        _,
        _,
    ) = token_network.functions.getChannelParticipantInfo(channel_identifier, B, A).call()
    assert B_deposit == 0
    assert B_withdrawn == 0
    assert B_is_the_closer == 0
    assert B_balance_hash == fake_bytes(32)
    assert B_nonce == 0


@pytest.fixture()
def update_state_tests(token_network: Contract, get_block: Callable) -> Callable:
    def get(
        channel_identifier: int,
        A: HexAddress,
        balance_proof_A: OnchainBalanceProof,
        B: HexAddress,
        balance_proof_B: OnchainBalanceProof,
        settle_timeout: int,
        txn_hash1: str,
    ) -> None:
        (settle_block_number, state) = token_network.functions.getChannelInfo(
            channel_identifier, A, B
        ).call()

        assert settle_block_number == settle_timeout + get_block(txn_hash1)
        assert state == ChannelState.CLOSED

        (
            _,
            _,
            A_is_the_closer,
            A_balance_hash,
            A_nonce,
            A_locksroot,
            A_locked,
        ) = token_network.functions.getChannelParticipantInfo(channel_identifier, A, B).call()
        assert A_is_the_closer is True
        assert A_balance_hash == balance_proof_A.balance_hash
        assert A_nonce == 5
        assert A_locksroot == NONEXISTENT_LOCKSROOT
        assert A_locked == 0

        (
            _,
            _,
            B_is_the_closer,
            B_balance_hash,
            B_nonce,
            B_locksroot,
            B_locked,
        ) = token_network.functions.getChannelParticipantInfo(channel_identifier, B, A).call()
        assert B_is_the_closer is False
        assert B_balance_hash == balance_proof_B.balance_hash
        assert B_nonce == 3
        assert B_locksroot == NONEXISTENT_LOCKSROOT
        assert B_locked == 0

    return get


@pytest.fixture()
def cooperative_settle_state_tests(token_network: Contract, custom_token: Contract) -> Callable:
    def get(
        channel_identifier: int,
        A: HexAddress,
        balance_A: int,
        B: HexAddress,
        balance_B: int,
        pre_account_balance_A: int,
        pre_account_balance_B: int,
        pre_balance_contract: int,
    ) -> None:
        common_settle_state_tests(
            custom_token,
            token_network,
            channel_identifier,
            A,
            balance_A,
            B,
            balance_B,
            pre_account_balance_A,
            pre_account_balance_B,
            pre_balance_contract,
        )

        (settle_block_number, state) = token_network.functions.getChannelInfo(
            channel_identifier, A, B
        ).call()
        assert settle_block_number == 0
        assert state == ChannelState.REMOVED

    return get


@pytest.fixture()
def settle_state_tests(token_network: Contract, custom_token: Contract) -> Callable:
    def get(
        channel_identifier: int,
        A: HexAddress,
        values_A: ChannelValues,
        B: HexAddress,
        values_B: ChannelValues,
        pre_account_balance_A: int,
        pre_account_balance_B: int,
        pre_balance_contract: int,
    ) -> None:
        # Calculate how much A and B should receive
        settlement = get_settlement_amounts(values_A, values_B)
        # Calculate how much A and B receive according to onchain computation
        on_chain_settlement = get_onchain_settlement_amounts(values_A, values_B)

        common_settle_state_tests(
            custom_token,
            token_network,
            channel_identifier,
            A,
            settlement.participant1_balance,
            B,
            settlement.participant2_balance,
            pre_account_balance_A,
            pre_account_balance_B,
            pre_balance_contract,
        )
        common_settle_state_tests(
            custom_token,
            token_network,
            channel_identifier,
            A,
            on_chain_settlement.participant1_balance,
            B,
            on_chain_settlement.participant2_balance,
            pre_account_balance_A,
            pre_account_balance_B,
            pre_balance_contract,
        )

        info_A = token_network.functions.getChannelParticipantInfo(channel_identifier, A, B).call()
        (_, _, _, _, _, locksroot_A, locked_amount_A) = info_A
        assert locked_amount_A == settlement.participant1_locked
        assert locked_amount_A == on_chain_settlement.participant1_locked
        if locked_amount_A > 0:
            assert locksroot_A == values_A.locksroot

        info_B = token_network.functions.getChannelParticipantInfo(channel_identifier, B, A).call()
        (_, _, _, _, _, locksroot_B, locked_amount_B) = info_B
        assert locked_amount_B == settlement.participant2_locked
        assert locked_amount_B == on_chain_settlement.participant2_locked
        if locked_amount_B > 0:
            assert locksroot_B == values_B.locksroot

        (settle_block_number, state) = token_network.functions.getChannelInfo(
            channel_identifier, A, B
        ).call()
        assert settle_block_number == 0
        if locked_amount_A > 0 or locked_amount_B > 0:
            assert state == ChannelState.SETTLED
        else:
            assert state == ChannelState.REMOVED

    return get


@pytest.fixture()
def unlock_state_tests(custom_token: Contract, token_network: Contract) -> Callable:
    def get(
        A: HexAddress,
        locked_A: int,
        locksroot_A: bytes,
        B: HexAddress,
        locked_B: int,
        pre_account_balance_A: int,
        pre_account_balance_B: int,
        pre_balance_contract: int,
    ) -> None:
        account_balance_A = custom_token.functions.balanceOf(A).call()
        account_balance_B = custom_token.functions.balanceOf(B).call()
        balance_contract = custom_token.functions.balanceOf(token_network.address).call()
        assert account_balance_A == pre_account_balance_A + locked_A
        assert account_balance_B == pre_account_balance_B + locked_B
        assert balance_contract == pre_balance_contract - locked_A - locked_B

        assert token_network.functions.getParticipantLockedAmount(A, B, locksroot_A).call() == 0

    return get


@pytest.fixture()
def withdraw_state_tests(custom_token: Contract, token_network: Contract) -> Callable:
    def get(
        channel_identifier: int,
        participant: HexAddress,
        deposit_participant: int,
        total_withdrawn_participant: int,
        pre_withdrawn_participant: int,
        pre_balance_participant: int,
        partner: HexAddress,
        deposit_partner: HexAddress,
        total_withdrawn_partner: int,
        pre_balance_partner: int,
        pre_balance_contract: int,
        delegate: Optional[HexAddress] = None,
    ) -> None:
        current_withdrawn_participant = total_withdrawn_participant - pre_withdrawn_participant
        (_, state) = token_network.functions.getChannelInfo(
            channel_identifier, participant, partner
        ).call()
        assert state == ChannelState.OPENED

        (
            deposit,
            withdrawn_amount,
            is_the_closer,
            balance_hash,
            nonce,
            locksroot,
            locked_amount,
        ) = token_network.functions.getChannelParticipantInfo(
            channel_identifier, participant, partner
        ).call()
        assert deposit == deposit_participant
        assert withdrawn_amount == total_withdrawn_participant
        assert is_the_closer is False
        assert balance_hash == fake_bytes(32)
        assert nonce == 0
        assert locksroot == NONEXISTENT_LOCKSROOT
        assert locked_amount == 0

        (
            deposit,
            withdrawn_amount,
            is_the_closer,
            balance_hash,
            nonce,
            locksroot,
            locked_amount,
        ) = token_network.functions.getChannelParticipantInfo(
            channel_identifier, partner, participant
        ).call()
        assert deposit == deposit_partner
        assert withdrawn_amount == total_withdrawn_partner
        assert is_the_closer is False
        assert balance_hash == fake_bytes(32)
        assert nonce == 0
        assert locksroot == NONEXISTENT_LOCKSROOT
        assert locked_amount == 0

        balance_participant = custom_token.functions.balanceOf(participant).call()
        balance_partner = custom_token.functions.balanceOf(partner).call()
        balance_contract = custom_token.functions.balanceOf(token_network.address).call()
        assert balance_participant == pre_balance_participant + current_withdrawn_participant
        assert balance_partner == pre_balance_partner
        assert balance_contract == pre_balance_contract - current_withdrawn_participant

        if delegate is not None:
            balance_delegate = custom_token.functions.balanceOf(delegate).call()
            assert balance_delegate == balance_delegate

    return get


@pytest.fixture(scope="session")
def create_balance_proof(token_network: Contract, get_private_key: Callable) -> Callable:
    def get(
        channel_identifier: ChannelID,
        participant: HexAddress,
        transferred_amount: TokenAmount = TokenAmount(0),  # noqa
        locked_amount: TokenAmount = TokenAmount(0),  # noqa
        nonce: Nonce = Nonce(0),  # noqa
        locksroot: Optional[Locksroot] = None,
        additional_hash: Optional[AdditionalHash] = None,
        v: int = 27,
        signer: Optional[HexAddress] = None,
        other_token_network: Optional[Contract] = None,
    ) -> OnchainBalanceProof:
        _token_network = other_token_network or token_network
        private_key = get_private_key(signer or participant)
        locksroot = locksroot or LOCKSROOT_OF_NO_LOCKS
        additional_hash = additional_hash or AdditionalHash(b"\x00" * 32)

        balance_hash = hash_balance_data(transferred_amount, locked_amount, locksroot)

        signature = sign_balance_proof(
            private_key,
            _token_network.address,
            _token_network.functions.chain_id().call(),
            channel_identifier,
            MessageTypeId.BALANCE_PROOF,
            balance_hash,
            nonce,
            additional_hash,
            v,
        )
        # The keys of the dictionary correspond to the parameters of
        # create_balance_proof_countersignature.
        return OnchainBalanceProof(
            balance_hash=balance_hash,
            nonce=nonce,
            additional_hash=additional_hash,
            original_signature=signature,
        )

    return get


@pytest.fixture(scope="session")
def create_balance_proof_countersignature(
    token_network: Contract, get_private_key: Callable
) -> Callable:
    def get(
        participant: HexAddress,
        channel_identifier: ChannelID,
        msg_type: MessageTypeId,
        balance_hash: BalanceHash,
        nonce: Nonce,
        additional_hash: AdditionalHash,
        original_signature: Signature,
        v: int = 27,
        other_token_network: Optional[Contract] = None,
    ) -> bytes:
        _token_network = other_token_network or token_network
        private_key = get_private_key(participant)

        non_closing_signature = sign_balance_proof_message(
            private_key,
            _token_network.address,
            _token_network.functions.chain_id().call(),
            channel_identifier,
            msg_type,
            balance_hash,
            nonce,
            additional_hash,
            original_signature,
            v,
        )
        return non_closing_signature

    return get


@pytest.fixture(scope="session")
def create_close_signature_for_no_balance_proof(
    token_network: Contract, get_private_key: Callable
) -> Callable:
    def get(
        participant: HexAddress,
        channel_identifier: ChannelID,
        v: int = 27,
        other_token_network: Optional[Contract] = None,
    ) -> bytes:
        _token_network = other_token_network or token_network
        private_key = get_private_key(participant)

        non_closing_signature = sign_balance_proof_message(
            private_key,
            _token_network.address,
            _token_network.functions.chain_id().call(),
            channel_identifier,
            MessageTypeId.BALANCE_PROOF,
            EMPTY_BALANCE_HASH,
            Nonce(0),
            EMPTY_ADDITIONAL_HASH,
            EMPTY_SIGNATURE,
            v,
        )
        return non_closing_signature

    return get


@pytest.fixture()
def create_withdraw_signatures(token_network: Contract, get_private_key: Callable) -> Callable:
    def get(
        participants_to_sign: Collection[HexAddress],
        channel_identifier: ChannelID,
        participant_who_withdraws: HexAddress,
        amount_to_withdraw: TokenAmount,
        expiration_block: BlockExpiration,
        token_network_address: Optional[HexAddress] = None,
        v: int = 27,
    ) -> List[bytes]:
        if token_network_address is None:
            token_network_address = token_network.address

        signatures = []
        for participant in participants_to_sign:
            private_key = get_private_key(participant)
            signature = sign_withdraw_message(
                private_key,
                token_network_address,
                token_network.functions.chain_id().call(),
                channel_identifier,
                participant_who_withdraws,
                amount_to_withdraw,
                expiration_block,
                v,
            )
            signatures.append(signature)
        return signatures

    return get


def call_settle(
    token_network: Contract,
    channel_identifier: int,
    A: HexAddress,
    vals_A: ChannelValues,
    B: HexAddress,
    vals_B: ChannelValues,
) -> None:
    A_total_transferred = vals_A.transferred + vals_A.locked_amounts.locked
    B_total_transferred = vals_B.transferred + vals_B.locked_amounts.locked
    assert B_total_transferred >= A_total_transferred

    if A_total_transferred != B_total_transferred:
        with pytest.raises(TransactionFailed):
            call_and_transact(
                token_network.functions.settleChannel(
                    channel_identifier=channel_identifier,
                    participant1=B,
                    participant1_transferred_amount=vals_B.transferred,
                    participant1_locked_amount=0,
                    participant1_locksroot=vals_B.locksroot,
                    participant2=A,
                    participant2_transferred_amount=0,
                    participant2_locked_amount=0,
                    participant2_locksroot=vals_A.locksroot,
                ),
                {"from": A},
            )

    contract_function = token_network.functions.settleChannel(
        channel_identifier,
        A,
        vals_A.transferred,
        vals_A.locked_amounts.locked,
        vals_A.locksroot,
        B,
        vals_B.transferred,
        vals_B.locked_amounts.locked,
        vals_B.locksroot,
    )
    # call() raises TransactionFailed exception
    contract_function.call({"from": A})
    # transact() changes the chain state
    call_and_transact(contract_function, {"from": A})



from typing import Callable

import pytest
from eth_typing.evm import HexAddress
from web3 import Web3
from web3.contract import Contract

from raiden_contracts.constants import CONTRACT_ONE_TO_N
from raiden_contracts.utils.proofs import sign_one_to_n_iou


@pytest.fixture(scope="session")
def one_to_n_contract(
    deploy_tester_contract: Callable,
    uninitialized_user_deposit_contract: Contract,
    web3: Web3,
    service_registry: Contract,
) -> Contract:
    chain_id = web3.eth.chain_id
    return deploy_tester_contract(
        CONTRACT_ONE_TO_N,
        _deposit_contract=uninitialized_user_deposit_contract.address,
        _chain_id=chain_id,
        _service_registry_contract=service_registry.address,
    )


@pytest.fixture(scope="session")
def one_to_n_internals(
    deploy_tester_contract: Callable,
    uninitialized_user_deposit_contract: Contract,
    web3: Web3,
    service_registry: Contract,
) -> Contract:
    chain_id = web3.eth.chain_id
    return deploy_tester_contract(
        "OneToNInternalsTest",
        _deposit_contract=uninitialized_user_deposit_contract.address,
        _chain_id=chain_id,
        _service_registry_contract=service_registry.address,
    )


@pytest.fixture
def make_iou(web3: Web3, one_to_n_contract: Contract, get_private_key: Callable) -> Callable:
    chain_id = web3.eth.chain_id

    def f(
        sender: HexAddress,
        receiver: HexAddress,
        amount: int = 10,
        expiration_block: int = None,
        chain_id: int = chain_id,
        one_to_n_address: HexAddress = one_to_n_contract.address,
    ) -> dict:
        if expiration_block is None:
            expiration_block = web3.eth.block_number + 10
        iou = dict(
            sender=sender,
            receiver=receiver,
            amount=amount,
            expiration_block=expiration_block,
            one_to_n_address=one_to_n_address,
            chain_id=chain_id,
        )
        iou["signature"] = sign_one_to_n_iou(get_private_key(sender), **iou)  # type: ignore
        del iou["chain_id"]
        del iou["one_to_n_address"]
        return iou

    return f



from typing import Callable, Dict

import pytest
from web3.contract import Contract

from raiden_contracts.constants import CONTRACT_DEPOSIT


@pytest.fixture
def get_deposit_contract(deploy_tester_contract: Callable) -> Callable:
    """Deploy a Deposit contract with the given arguments"""

    def get(**arguments: Dict) -> Contract:
        return deploy_tester_contract(CONTRACT_DEPOSIT, **arguments)

    return get



from typing import Any, Callable, Dict, List

from eth_typing.evm import HexAddress


def check_secret_revealed(secrethash: bytes, secret: bytes) -> Callable[..., Any]:
    def get(event: Any) -> Any:
        assert event["args"]["secrethash"] == secrethash
        assert event["args"]["secret"] == secret

    return get


def check_secrets_revealed(secrethashes: List[bytes], secrets: List[bytes]) -> Callable[..., Any]:
    def get(event: Any) -> Any:
        assert event["args"]["secrethash"] in secrethashes
        assert event["args"]["secret"] in secrets

    return get


def check_token_network_created(
    token_address: HexAddress, token_network_address: HexAddress
) -> Callable[..., None]:
    def get(event: Dict[str, Any]) -> None:
        assert event["args"]["token_address"] == token_address
        assert event["args"]["token_network_address"] == token_network_address

    return get


def check_address_registered(eth_address: HexAddress, endpoint: str) -> Callable[..., Any]:
    def get(event: Dict[str, Any]) -> None:
        assert event["args"]["eth_address"] == eth_address
        assert event["args"]["endpoint"] == endpoint

    return get


def check_channel_opened(
    channel_identifier: int,
    participant1: HexAddress,
    participant2: HexAddress,
    settle_timeout: int,
) -> Callable[[Dict[str, Any]], None]:
    def get(event: Dict[str, Any]) -> None:
        assert event["args"]["channel_identifier"] == channel_identifier
        assert event["args"]["participant1"] == participant1
        assert event["args"]["participant2"] == participant2
        assert event["args"]["settle_timeout"] == settle_timeout

    return get


# Check TokenNetwork.ChannelNewDeposit events. Not for UDC deposits!
def check_new_deposit(
    channel_identifier: int, participant: HexAddress, deposit: int
) -> Callable[[Dict[str, Any]], None]:
    def get(event: Dict[str, Any]) -> None:
        assert event["args"]["channel_identifier"] == channel_identifier
        assert event["args"]["participant"] == participant
        assert event["args"]["total_deposit"] == deposit

    return get


def check_withdraw(
    channel_identifier: int, participant: HexAddress, withdrawn_amount: int
) -> Callable[[Dict[str, Any]], None]:
    def get(event: Dict[str, Any]) -> None:
        assert event["args"]["channel_identifier"] == channel_identifier
        assert event["args"]["participant"] == participant
        assert event["args"]["total_withdraw"] == withdrawn_amount

    return get


def check_withdraw_2(
    channel_identifier: int, participant: HexAddress, withdrawn_amount: int
) -> Callable[[Dict[str, Any]], bool]:
    def get(event: Dict[str, Any]) -> bool:
        is_match = (
            event["args"]["channel_identifier"] == channel_identifier
            and event["args"]["participant"] == participant
            and event["args"]["total_withdraw"] == withdrawn_amount
        )
        return is_match

    return get


def check_channel_closed(
    channel_identifier: int, closing_participant: HexAddress, nonce: int, balance_hash: bytes
) -> Callable[[Dict[str, Any]], None]:
    def get(event: Dict[str, Any]) -> None:
        assert event["args"]["channel_identifier"] == channel_identifier
        assert event["args"]["closing_participant"] == closing_participant
        assert event["args"]["nonce"] == nonce
        assert event["args"]["balance_hash"] == balance_hash

    return get


def check_channel_unlocked(
    channel_identifier: int,
    receiver: HexAddress,
    sender: HexAddress,
    locksroot: bytes,
    unlocked_amount: int,
    returned_tokens: int,
) -> Callable[[Dict[str, Any]], None]:
    def get(event: Dict[str, Any]) -> None:
        assert event["args"]["channel_identifier"] == channel_identifier
        assert event["args"]["receiver"] == receiver
        assert event["args"]["sender"] == sender
        assert event["args"]["locksroot"] == locksroot
        assert event["args"]["unlocked_amount"] == unlocked_amount
        assert event["args"]["returned_tokens"] == returned_tokens

    return get


def check_transfer_updated(
    channel_identifier: int, closing_participant: HexAddress, nonce: int, balance_hash: bytes
) -> Callable[[Dict[str, Any]], None]:
    def get(event: Dict[str, Any]) -> None:
        assert event["args"]["channel_identifier"] == channel_identifier
        assert event["args"]["closing_participant"] == closing_participant
        assert event["args"]["nonce"] == nonce
        assert event["args"]["balance_hash"] == balance_hash

    return get


def check_channel_settled(
    channel_identifier: int,
    participant1: HexAddress,
    participant1_amount: int,
    participant1_locksroot: bytes,
    participant2: HexAddress,
    participant2_amount: int,
    participant2_locksroot: bytes,
) -> Callable[[Dict[str, Any]], None]:
    def get(event: Dict[str, Any]) -> None:
        assert event["args"]["channel_identifier"] == channel_identifier
        assert event["args"]["participant1"] == participant1
        assert event["args"]["participant1_amount"] == participant1_amount
        assert event["args"]["participant1_locksroot"] == participant1_locksroot
        assert event["args"]["participant2"] == participant2
        assert event["args"]["participant2_amount"] == participant2_amount
        assert event["args"]["participant2_locksroot"] == participant2_locksroot

    return get



from typing import Callable, Dict

import pytest
from eth_typing import HexAddress
from web3 import Web3
from web3.contract import Contract, get_event_data

from raiden_contracts.constants import (
    CONTRACT_TOKEN_NETWORK,
    CONTRACT_TOKEN_NETWORK_REGISTRY,
    EVENT_TOKEN_NETWORK_CREATED,
    MAX_ETH_CHANNEL_PARTICIPANT,
    MAX_ETH_TOKEN_NETWORK,
    TEST_SETTLE_TIMEOUT_MAX,
    TEST_SETTLE_TIMEOUT_MIN,
)
from raiden_contracts.contract_manager import ContractManager
from raiden_contracts.tests.utils import call_and_transact
from raiden_contracts.tests.utils.constants import DEPLOYER_ADDRESS

snapshot_before_token_network = None


@pytest.fixture
def get_token_network(deploy_tester_contract: Callable) -> Callable:
    """Deploy a token network as a separate contract (registry is not used)"""

    def get(**arguments: Dict) -> Contract:
        return deploy_tester_contract(CONTRACT_TOKEN_NETWORK, **arguments)

    return get


@pytest.fixture(scope="session")
def register_token_network(web3: Web3, contracts_manager: ContractManager) -> Callable:
    """Returns a function that uses token_network_registry fixture to register
    and deploy a new token network"""

    def get(
        token_network_registry: Contract,
        token_address: HexAddress,
        channel_participant_deposit_limit: int,
        token_network_deposit_limit: int,
    ) -> Contract:
        tx_hash = call_and_transact(
            token_network_registry.functions.createERC20TokenNetwork(
                token_address,
                channel_participant_deposit_limit,
                token_network_deposit_limit,
            ),
            {"from": DEPLOYER_ADDRESS},
        )
        tx_receipt = web3.eth.get_transaction_receipt(tx_hash)
        event_abi = contracts_manager.get_event_abi(
            CONTRACT_TOKEN_NETWORK_REGISTRY, EVENT_TOKEN_NETWORK_CREATED
        )
        event_data = get_event_data(
            abi_codec=web3.codec, event_abi=event_abi, log_entry=tx_receipt["logs"][0]
        )
        contract_address = event_data["args"]["token_network_address"]
        contract = web3.eth.contract(
            abi=contracts_manager.get_contract_abi(CONTRACT_TOKEN_NETWORK),
            address=contract_address,
        )
        return contract

    return get


@pytest.fixture(scope="session")
def channel_participant_deposit_limit() -> int:
    return MAX_ETH_CHANNEL_PARTICIPANT


@pytest.fixture(scope="session")
def token_network_deposit_limit() -> int:
    return MAX_ETH_TOKEN_NETWORK


@pytest.fixture
def no_token_network(web3: Web3) -> None:
    """Some tests must be executed before a token network gets created

    These tests should use this fixture. Otherwise a session level token
    network might already be registered.
    """
    if snapshot_before_token_network is not None:
        web3.testing.revert(snapshot_before_token_network)


@pytest.fixture(scope="session")
def token_network(
    register_token_network: Callable,
    custom_token: Contract,
    channel_participant_deposit_limit: int,
    token_network_deposit_limit: int,
    token_network_registry_contract: Contract,
    web3: Web3,
) -> Contract:
    """Register a new token network for a custom token"""
    global snapshot_before_token_network
    snapshot_before_token_network = web3.testing.snapshot()  # type: ignore
    return register_token_network(
        token_network_registry_contract,
        custom_token.address,
        channel_participant_deposit_limit,
        token_network_deposit_limit,
    )


@pytest.fixture(scope="session")
def token_network_in_another_token_network_registry(
    register_token_network: Callable,
    custom_token: Contract,
    channel_participant_deposit_limit: int,
    token_network_deposit_limit: int,
    token_network_registry_contract2: Contract,
) -> Contract:
    """TokenNetwork registered to a foreign TokenNetworkRegistry

    with which the service payments should not work."""
    return register_token_network(
        token_network_registry_contract2,
        custom_token.address,
        channel_participant_deposit_limit,
        token_network_deposit_limit,
    )


@pytest.fixture
def token_network_contract(
    deploy_tester_contract: Callable,
    secret_registry_contract: Contract,
    standard_token_contract: Contract,
    web3: Web3,
) -> Contract:
    return deploy_tester_contract(
        CONTRACT_TOKEN_NETWORK,
        [
            standard_token_contract.address,
            secret_registry_contract.address,
            web3.eth.chain_id,
        ],
    )


@pytest.fixture()
def token_network_external(
    web3: Web3,
    get_token_network: Callable,
    custom_token: Contract,
    secret_registry_contract: Contract,
    channel_participant_deposit_limit: int,
    token_network_deposit_limit: int,
) -> Contract:
    return get_token_network(
        _token_address=custom_token.address,
        _secret_registry=secret_registry_contract.address,
        _chain_id=web3.eth.chain_id,
        _settlement_timeout_min=TEST_SETTLE_TIMEOUT_MIN,
        _settlement_timeout_max=TEST_SETTLE_TIMEOUT_MAX,
        _controller=DEPLOYER_ADDRESS,
        _channel_participant_deposit_limit=channel_participant_deposit_limit,
        _token_network_deposit_limit=token_network_deposit_limit,
    )



from typing import Callable

import pytest
from web3.contract import Contract

from raiden_contracts.constants import CONTRACT_SECRET_REGISTRY


@pytest.fixture(scope="session")
def secret_registry_contract(deploy_tester_contract: Callable) -> Contract:
    """Deployed SecretRegistry contract"""
    return deploy_tester_contract(CONTRACT_SECRET_REGISTRY)



"""
A simple Python script to deploy compiled contracts.
"""
import functools
import json
import logging
from logging import getLogger
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Union

import click
from click import BadParameter, Context, IntRange, Option, Parameter
from eth_typing import URI
from eth_typing.evm import ChecksumAddress, HexAddress
from eth_utils import is_address, to_checksum_address
from web3 import HTTPProvider, Web3
from web3.middleware import geth_poa_middleware

from raiden_contracts.constants import (
    CONTRACT_CUSTOM_TOKEN,
    CONTRACT_TOKEN_NETWORK_REGISTRY,
    DEPLOY_SETTLE_TIMEOUT_MAX,
    DEPLOY_SETTLE_TIMEOUT_MIN,
)
from raiden_contracts.contract_manager import DeployedContracts, contracts_deployed_path
from raiden_contracts.deploy.contract_deployer import ContractDeployer
from raiden_contracts.deploy.contract_verifier import ContractVerifier
from raiden_contracts.utils.private_key import get_private_key
from raiden_contracts.utils.signature import private_key_to_address
from raiden_contracts.utils.type_aliases import ChainID
from raiden_contracts.utils.versions import contracts_version_with_max_token_networks

LOG = getLogger(__name__)


def validate_address(
    _: Context, _param: Union[Option, Parameter], value: Optional[str]
) -> Optional[ChecksumAddress]:
    if not value:
        return None
    try:
        is_address(value)
        return to_checksum_address(value)
    except ValueError:
        raise click.BadParameter("must be a valid ethereum address")


def error_removed_option(message: str) -> Callable:
    """Takes a message and returns a callback that raises NoSuchOption

    if the value is not None. The message is used as an argument to NoSuchOption."""

    def f(_: Any, param: Parameter, value: Any) -> None:
        if value is not None:
            param_name = param.name.replace("_", "-") if param.name else "unnamed-param"
            raise click.NoSuchOption(f"--{param_name} is no longer a valid option. " + message)

    return f


def common_options(func: Callable) -> Callable:
    """A decorator that combines commonly appearing @click.option decorators."""

    @click.option("--private-key", required=True, help="Path to a private key store.")
    @click.option(
        "--password-file",
        help="Text file containing the password for the provided account",
        default=None,
        type=click.Path(exists=True, dir_okay=False),
        show_default=True,
        callback=lambda ctx, param, value: Path(value) if value is not None else None,
    )
    @click.option(
        "--rpc-provider",
        default="http://127.0.0.1:8545",
        help="Address of the Ethereum RPC provider",
    )
    @click.option("--wait", default=300, help="Max tx wait time in s.")
    @click.option("--gas-price", default=5, type=IntRange(min=1), help="Gas price to use in gwei")
    @click.option("--gas-limit", default=5_500_000)
    @click.option(
        "--contracts-version",
        default=None,
        help="Contracts version to verify. Current version will be used by default.",
    )
    @functools.wraps(func)
    def wrapper(*args: List, **kwargs: Dict) -> Any:
        return func(*args, **kwargs)

    return wrapper


# pylint: disable=R0913
def setup_ctx(
    ctx: click.Context,
    private_key: str,
    password_file: Optional[Path],
    rpc_provider: URI,
    wait: int,
    gas_price: int,
    gas_limit: int,
    contracts_version: Optional[str] = None,
) -> None:
    """Set up deployment context according to common options (shared among all
    subcommands).
    """

    logging.basicConfig(level=logging.DEBUG)
    logging.getLogger("web3").setLevel(logging.INFO)
    logging.getLogger("urllib3").setLevel(logging.INFO)

    web3 = Web3(HTTPProvider(rpc_provider, request_kwargs={"timeout": 60}))
    web3.middleware_onion.inject(geth_poa_middleware, layer=0)
    print("Web3 provider is", web3.provider)
    private_key_string = get_private_key(Path(private_key).expanduser(), password_file)
    if not private_key_string:
        raise RuntimeError("Could not access the private key.")
    owner = private_key_to_address(private_key_string)
    # pylint: disable=E1101
    if web3.eth.get_balance(owner) == 0:
        raise RuntimeError("Account with insufficient funds.")
    deployer = ContractDeployer(
        web3=web3,
        private_key=private_key_string,
        gas_limit=gas_limit,
        gas_price=gas_price,
        wait=wait,
        contracts_version=contracts_version,
    )
    ctx.obj = {
        "deployer": deployer,
        "deployed_contracts": {},
        "token_type": "CustomToken",
        "wait": wait,
    }


@click.group(chain=True)
def main() -> int:
    pass


def check_version_dependent_parameters(
    contracts_version: Optional[str], max_token_networks: Optional[int]
) -> None:
    required = contracts_version_with_max_token_networks(contracts_version)
    got = max_token_networks is not None

    # For newer conracts --max-token-networks is necessary.
    if required and not got:
        raise BadParameter(
            f"For contracts_version {contracts_version},"
            " --max-token-networks option is necessary.  See --help."
        )
    # For older contracts --max_token_networks is forbidden.
    if not required and got:
        raise BadParameter(
            f"For contracts_version {contracts_version},"
            " --max-token-networks option is forbidden"
            " because TokenNetworkRegistry this version is not configurable this way."
        )


@main.command()
@common_options
@click.option("--save-info/--no-save-info", default=True, help="Save deployment info to a file.")
@click.option(
    "--max-token-networks",
    help="The maximum number of tokens that can be registered.",
    type=int,
)
@click.option(
    "--secret-registry-from-deployment-file",
    type=click.Path(exists=True),
    help="The deployment file from which SecretRegistry should be reused",
)
@click.option(
    "--settle-timeout-max",
    type=click.INT,
    default=DEPLOY_SETTLE_TIMEOUT_MAX,
    help=(
        "The maximum allowed settle timeout for the channels opened with this "
        "set of smart contracts"
    ),
)
@click.option(
    "--settle-timeout-min",
    type=click.INT,
    default=DEPLOY_SETTLE_TIMEOUT_MIN,
    help=(
        "The minimum allowed settle timeout for the channels opened with this "
        "set of smart contracts"
    ),
)
@click.pass_context
def raiden(
    ctx: click.Context,
    private_key: str,
    password_file: Optional[Path],
    rpc_provider: URI,
    wait: int,
    gas_price: int,
    gas_limit: int,
    save_info: int,
    settle_timeout_min: int,
    settle_timeout_max: int,
    contracts_version: Optional[str],
    max_token_networks: Optional[int],
    secret_registry_from_deployment_file: Optional[str],
) -> None:
    check_version_dependent_parameters(contracts_version, max_token_networks)
    secret_registry_from_deployment_path: Optional[Path] = None
    if secret_registry_from_deployment_file:
        secret_registry_from_deployment_path = Path(secret_registry_from_deployment_file)

    setup_ctx(
        ctx=ctx,
        private_key=private_key,
        password_file=password_file,
        rpc_provider=rpc_provider,
        wait=wait,
        gas_price=gas_price,
        gas_limit=gas_limit,
        contracts_version=contracts_version,
    )
    deployer = ctx.obj["deployer"]
    deployed_contracts_info = deployer.deploy_raiden_contracts(
        max_num_of_token_networks=max_token_networks,
        reuse_secret_registry_from_deploy_file=secret_registry_from_deployment_path,
        settle_timeout_min=settle_timeout_min,
        settle_timeout_max=settle_timeout_max,
    )
    deployed_contracts = {
        contract_name: info["address"]
        for contract_name, info in deployed_contracts_info["contracts"].items()
    }

    if save_info:
        deployer.store_and_verify_deployment_info_raiden(
            deployed_contracts_info=deployed_contracts_info
        )
    else:
        deployer.verify_deployment_data(deployed_contracts_info=deployed_contracts_info)

    print(json.dumps(deployed_contracts, indent=4))
    ctx.obj["deployed_contracts"].update(deployed_contracts)


@main.command()
@common_options
@click.option(
    "--token-address",
    default="0x255Aa6DF07540Cb5d3d297f0D0D4D84cb52bc8e6",
    callback=validate_address,
    help="Address of token used to pay for the services (MS, PFS).",
)
@click.option(
    "--user-deposit-whole-limit",
    required=True,
    type=int,
    help="Maximum amount of tokens deposited in UserDeposit",
)
@click.option(
    "--service-registry-controller",
    required=True,
    callback=validate_address,
    help="Address of the controller that can modify the parameters of ServiceRegistry",
)
@click.option(
    "--service-deposit-bump-numerator",
    type=int,
    required=True,
    help="The numerator of the deposit bump after somebody makes a deposit into ServiceRegistry",
)
@click.option(
    "--service-deposit-bump-denominator",
    required=True,
    type=int,
    help="The denominator of the deposit bump after somebody makes a deposit into ServiceRegistry",
)
@click.option(
    "--service-deposit-decay-constant",
    required=True,
    type=int,
    help="The number of seconds for the price of ServiceRegistry to become roughly 1 / 2.7 of "
    "the original, if deposits are made.",
)
@click.option(
    "--initial-service-deposit-price",
    required=True,
    type=int,
    help="Initial amount of deposit for a registration in ServiceRegistry",
)
@click.option(
    "--service-deposit-min-price",
    required=True,
    type=int,
    help="The minimum amount of deposits, where the decay of the price stops.",
)
@click.option(
    "--service-registration-duration",
    required=True,
    type=int,
    help="The duration of service registration (seconds)",
)
@click.option("--save-info/--no-save-info", default=True, help="Save deployment info to a file.")
@click.option(
    "--token-network-registry-address",
    required=True,
    callback=validate_address,
    help="Address of TokenNetworkRegistry that MS contract looks at",
)
@click.option(
    "--service-registry-from-deployment-file",
    type=click.Path(exists=True),
    help="The deployment file from which ServiceRegistry should be reused",
)
@click.pass_context
def services(
    ctx: Context,
    private_key: str,
    password_file: Optional[Path],
    rpc_provider: URI,
    wait: int,
    gas_price: int,
    gas_limit: int,
    token_address: HexAddress,
    save_info: bool,
    contracts_version: Optional[str],
    user_deposit_whole_limit: int,
    service_registry_controller: HexAddress,
    initial_service_deposit_price: int,
    service_deposit_bump_numerator: int,
    service_deposit_bump_denominator: int,
    service_deposit_decay_constant: int,
    service_deposit_min_price: int,
    service_registration_duration: int,
    token_network_registry_address: HexAddress,
    service_registry_from_deployment_file: Optional[str],
) -> None:
    service_registry_from_deployment_path: Optional[Path] = None
    if service_registry_from_deployment_file:
        service_registry_from_deployment_path = Path(service_registry_from_deployment_file)

    setup_ctx(
        ctx=ctx,
        private_key=private_key,
        password_file=password_file,
        rpc_provider=rpc_provider,
        wait=wait,
        gas_price=gas_price,
        gas_limit=gas_limit,
        contracts_version=contracts_version,
    )
    deployer: ContractDeployer = ctx.obj["deployer"]

    deployed_contracts_info = deployer.deploy_service_contracts(
        token_address=token_address,
        user_deposit_whole_balance_limit=user_deposit_whole_limit,
        service_registry_controller=service_registry_controller,
        initial_service_deposit_price=initial_service_deposit_price,
        service_deposit_bump_numerator=service_deposit_bump_numerator,
        service_deposit_bump_denominator=service_deposit_bump_denominator,
        decay_constant=service_deposit_decay_constant,
        min_price=service_deposit_min_price,
        registration_duration=service_registration_duration,
        token_network_registry_address=token_network_registry_address,
        reuse_service_registry_from_deploy_file=service_registry_from_deployment_path,
    )
    deployed_contracts = {
        contract_name: info["address"]
        for contract_name, info in deployed_contracts_info["contracts"].items()
    }

    if save_info:
        deployer.store_and_verify_deployment_info_services(
            deployed_contracts_info=deployed_contracts_info,
            token_address=token_address,
            user_deposit_whole_balance_limit=user_deposit_whole_limit,
            token_network_registry_address=token_network_registry_address,
        )
    else:
        deployer.verify_service_contracts_deployment_data(
            deployed_contracts_info=deployed_contracts_info,
            token_address=token_address,
            user_deposit_whole_balance_limit=user_deposit_whole_limit,
            token_network_registry_address=token_network_registry_address,
        )

    print(json.dumps(deployed_contracts, indent=4))
    ctx.obj["deployed_contracts"].update(deployed_contracts)


@main.command()
@common_options
@click.option(
    "--token-supply",
    default=10000000,
    help="Token contract supply (number of total issued tokens).",
)
@click.option("--token-name", default=CONTRACT_CUSTOM_TOKEN, help="Token contract name.")
@click.option("--token-decimals", default=18, help="Token contract number of decimals.")
@click.option("--token-symbol", default="TKN", help="Token contract symbol.")
@click.pass_context
def token(
    ctx: click.Context,
    private_key: str,
    password_file: Optional[Path],
    rpc_provider: URI,
    wait: int,
    gas_price: int,
    gas_limit: int,
    contracts_version: Optional[str],
    token_supply: int,
    token_name: str,
    token_decimals: int,
    token_symbol: str,
) -> None:
    setup_ctx(
        ctx=ctx,
        private_key=private_key,
        password_file=password_file,
        rpc_provider=rpc_provider,
        wait=wait,
        gas_price=gas_price,
        gas_limit=gas_limit,
        contracts_version=contracts_version,
    )
    deployer = ctx.obj["deployer"]
    token_supply *= 10 ** token_decimals
    deployed_token = deployer.deploy_token_contract(
        token_supply,
        token_decimals,
        token_name,
        token_symbol,
        token_type=ctx.obj["token_type"],
    )
    print(json.dumps(deployed_token, indent=4))
    ctx.obj["deployed_contracts"].update(deployed_token)


@main.command()
@common_options
@click.option(
    "--token-address",
    default=None,
    callback=validate_address,
    help="Already deployed token address.",
)
@click.option(
    "--registry-address",
    default=None,
    callback=error_removed_option("Use --token-network-registry-address"),
    hidden=True,
    help="Renamed into --token-network-registry-address",
)
@click.option(
    "--token-network-registry-address",
    default=None,
    callback=validate_address,
    help="Address of token network registry",
)
@click.option(
    "--channel-participant-deposit-limit",
    type=int,
    help="Address of token network registry",
)
@click.option(
    "--token-network-deposit-limit",
    type=int,
    help="Address of token network registry",
)
@click.pass_context
def register(
    ctx: Context,
    private_key: str,
    password_file: Optional[Path],
    rpc_provider: URI,
    wait: int,
    gas_price: int,
    gas_limit: int,
    contracts_version: str,
    token_address: HexAddress,
    token_network_registry_address: HexAddress,
    channel_participant_deposit_limit: int,
    token_network_deposit_limit: int,
    registry_address: Optional[HexAddress],
) -> None:
    assert registry_address is None  # No longer used option
    setup_ctx(
        ctx=ctx,
        private_key=private_key,
        password_file=password_file,
        rpc_provider=rpc_provider,
        wait=wait,
        gas_price=gas_price,
        gas_limit=gas_limit,
        contracts_version=contracts_version,
    )
    token_type = ctx.obj["token_type"]
    deployer = ctx.obj["deployer"]

    if token_address:
        ctx.obj["deployed_contracts"][token_type] = token_address
    if token_network_registry_address:
        ctx.obj["deployed_contracts"][
            CONTRACT_TOKEN_NETWORK_REGISTRY
        ] = token_network_registry_address

    if CONTRACT_TOKEN_NETWORK_REGISTRY not in ctx.obj["deployed_contracts"]:
        raise RuntimeError(
            "No TokenNetworkRegistry was specified. "
            "Add --token-network-registry-address <address>."
        )
    assert token_type in ctx.obj["deployed_contracts"]
    abi = deployer.contract_manager.get_contract_abi(CONTRACT_TOKEN_NETWORK_REGISTRY)
    token_network = deployer.register_token_network(
        token_registry_abi=abi,
        token_registry_address=ctx.obj["deployed_contracts"][CONTRACT_TOKEN_NETWORK_REGISTRY],
        token_address=ctx.obj["deployed_contracts"][token_type],
        channel_participant_deposit_limit=channel_participant_deposit_limit,
        token_network_deposit_limit=token_network_deposit_limit,
    )
    _add_token_network_deploy_info(token_network, deployer, contracts_version)


def _add_token_network_deploy_info(
    token_network: Dict[str, Any], deployer: ContractDeployer, contracts_version: str
) -> None:
    """Add deploy info dict to the deploy_*.json file"""
    deployment_file_path = contracts_deployed_path(
        chain_id=ChainID(deployer.web3.eth.chain_id), version=contracts_version
    )
    with deployment_file_path.open() as f:
        deployed_contracts_info: DeployedContracts = json.load(f)
    deployed_contracts_info.setdefault("token_networks", []).append(token_network)
    deployer.store_and_verify_deployment_info_raiden(
        deployed_contracts_info=deployed_contracts_info
    )


@main.command()
@click.option(
    "--rpc-provider",
    default="http://127.0.0.1:8545",
    help="Address of the Ethereum RPC provider",
)
@click.option(
    "--contracts-version",
    help="Contracts version to verify. Current version will be used by default.",
)
@click.pass_context
def verify(_: Any, rpc_provider: URI, contracts_version: Optional[str]) -> None:
    web3 = Web3(HTTPProvider(rpc_provider, request_kwargs={"timeout": 60}))
    web3.middleware_onion.inject(geth_poa_middleware, layer=0)
    print("Web3 provider is", web3.provider)

    verifier = ContractVerifier(web3=web3, contracts_version=contracts_version)
    verifier.verify_deployed_contracts_in_filesystem()


if __name__ == "__main__":
    main()



from typing import Callable

from web3.contract import Contract

from raiden_contracts.constants import TEST_SETTLE_TIMEOUT_MIN, ChannelEvent
from raiden_contracts.tests.utils import call_and_transact
from raiden_contracts.utils.events import check_channel_opened, check_new_deposit


def test_channel_open_with_deposit_basics(
    get_accounts: Callable,
    token_network: Contract,
    assign_tokens: Callable,
) -> None:
    """Some basic checks that `open_withDeposit` works.

    Detailed tests exist for `openChannel` and `setTotalDeposit`, and as `openChannelWithDeposit`
    is a simple wrapper, these don't need to be duplicated.
    """
    (A, B, C, D) = get_accounts(4)
    deposit = 100

    # Check channel creation by participant
    assign_tokens(A, deposit)
    call_and_transact(
        token_network.functions.openChannelWithDeposit(A, B, TEST_SETTLE_TIMEOUT_MIN, deposit),
        {"from": A},
    )
    assert token_network.functions.getChannelIdentifier(A, B).call() == 1
    assert token_network.functions.getChannelParticipantInfo(1, A, B).call()[0] == 100
    assert token_network.functions.getChannelParticipantInfo(1, B, A).call()[0] == 0

    # Check channel creation by delegate
    assign_tokens(D, deposit)
    call_and_transact(
        token_network.functions.openChannelWithDeposit(B, C, TEST_SETTLE_TIMEOUT_MIN, deposit),
        {"from": D},
    )
    assert token_network.functions.getChannelIdentifier(B, C).call() == 2
    assert token_network.functions.getChannelParticipantInfo(2, B, C).call()[0] == 100
    assert token_network.functions.getChannelParticipantInfo(2, C, B).call()[0] == 0


def test_channel_open_with_deposit_events(
    get_accounts: Callable,
    token_network: Contract,
    event_handler: Callable,
    assign_tokens: Callable,
) -> None:
    """A successful openChannelWithDeposit() causes an OPENED and DEPOSIT event"""
    ev_handler = event_handler(token_network)
    (A, B) = get_accounts(2)
    deposit = 100

    assign_tokens(A, deposit)
    txn_hash = call_and_transact(
        token_network.functions.openChannelWithDeposit(A, B, TEST_SETTLE_TIMEOUT_MIN, deposit),
        {"from": A},
    )
    channel_identifier = token_network.functions.getChannelIdentifier(A, B).call()

    ev_handler.add(
        txn_hash,
        ChannelEvent.OPENED,
        check_channel_opened(channel_identifier, A, B, TEST_SETTLE_TIMEOUT_MIN),
    )
    ev_handler.add(
        txn_hash,
        ChannelEvent.DEPOSIT,
        check_new_deposit(channel_identifier, A, deposit),
    )
    ev_handler.check()



#!/usr/bin/env python
import json
import re
import sys
from typing import Dict, List, Set, TextIO

import click
from click.types import File

IMPORT_RE = re.compile(r'^import +["\'](?P<contract>[^"\']+.sol)["\'];$')

"""
Utility to join solidity contracts into a single output file by recursively
resolving imports.

example usage:

$ cd raiden-contracts/raiden_contracts
$ python ./utils/join_contracts.py ./contracts/TokenNetwork.sol joined.sol

"""


class ContractJoiner:
    def __init__(self, import_map: Dict[str, str] = None):
        self.seen_pragmas: Set[str] = set()
        self.seen_contracts: Set[str] = set()
        self.import_map = import_map if import_map else {}

    def join(self, contract_file: TextIO) -> List[str]:
        out: List[str] = []
        if contract_file.name in self.seen_contracts:
            print("Skipping duplicate {}".format(contract_file.name), file=sys.stderr)
            return []

        self.seen_contracts.add(contract_file.name)
        print("Reading {}".format(contract_file.name), file=sys.stderr)

        for line in contract_file:
            line = line.strip("\r\n")
            stripped_line = line.strip()
            if stripped_line.startswith("pragma") or stripped_line.startswith(
                "// SPDX-License-Identifier: MIT"
            ):
                self._on_pragma_line(line=line, out=out)
            elif stripped_line.startswith("import"):
                self._on_import_line(stripped_line=stripped_line, out=out)
            else:
                out.append(line)
        return out

    def _on_pragma_line(self, line: str, out: List[str]) -> None:
        if line not in self.seen_pragmas:
            self.seen_pragmas.add(line)
            out.append(line)

    def _on_import_line(self, stripped_line: str, out: List[str]) -> None:
        match = IMPORT_RE.match(stripped_line)
        if match:
            next_file = match.groupdict().get("contract")
            assert next_file
            for prefix, path in self.import_map.items():
                if next_file.startswith(prefix):
                    next_file = next_file.replace(prefix, path)
            with open(next_file) as next_contract:
                out.extend(self.join(next_contract))


@click.command()
@click.option(
    "--import-map", default="{}", help='JSON mapping {"path-prefix": "/file/system/path"}'
)
@click.argument("contract", type=File())
@click.argument("output", type=File("w"))
def main(contract: TextIO, output: TextIO, import_map: str) -> None:
    import_map_json = json.loads(import_map)
    output.write("\n".join(ContractJoiner(import_map_json).join(contract)))


if __name__ == "__main__":
    main()  # pylint: disable=E1120



from typing import Callable

import pytest
from eth_tester.exceptions import TransactionFailed
from web3 import Web3
from web3.contract import Contract

from raiden_contracts.constants import TEST_SETTLE_TIMEOUT_MAX, TEST_SETTLE_TIMEOUT_MIN
from raiden_contracts.tests.utils import LOCKSROOT_OF_NO_LOCKS, call_and_transact, fake_bytes
from raiden_contracts.tests.utils.blockchain import mine_blocks


def test_settle_timeout_inrange(
    token_network: Contract,
    get_accounts: Callable,
    web3: Web3,
    create_close_signature_for_no_balance_proof: Callable,
) -> None:
    """The TokenNetwork constructor must enforce that settle timeout is in
    the valid range.

    Also asserts that the constants.py and the netting channel contract values
    are synched.
    """
    (A, B) = get_accounts(2)

    small_settle_timeout = TEST_SETTLE_TIMEOUT_MIN - 1
    large_settle_timeout = TEST_SETTLE_TIMEOUT_MAX + 1

    with pytest.raises(TransactionFailed, match="TN: settle timeout < min"):
        token_network.functions.openChannel(A, B, small_settle_timeout).call()

    with pytest.raises(TransactionFailed, match="TN: settle timeout > max"):
        token_network.functions.openChannel(A, B, large_settle_timeout).call()

    call_and_transact(token_network.functions.openChannel(A, B, TEST_SETTLE_TIMEOUT_MIN))
    channel_identifier = token_network.functions.getChannelIdentifier(A, B).call()
    (settle_block_number, _) = token_network.functions.getChannelInfo(
        channel_identifier, A, B
    ).call()

    assert settle_block_number == TEST_SETTLE_TIMEOUT_MIN

    closing_sig = create_close_signature_for_no_balance_proof(A, channel_identifier)
    call_and_transact(
        token_network.functions.closeChannel(
            channel_identifier=channel_identifier,
            non_closing_participant=B,
            closing_participant=A,
            balance_hash=fake_bytes(32),
            nonce=0,
            additional_hash=fake_bytes(32),
            non_closing_signature=fake_bytes(65),
            closing_signature=closing_sig,
        ),
        {"from": A},
    )
    mine_blocks(web3, TEST_SETTLE_TIMEOUT_MIN + 1)
    call_and_transact(
        token_network.functions.settleChannel(
            channel_identifier,
            A,
            0,
            0,
            LOCKSROOT_OF_NO_LOCKS,
            B,
            0,
            0,
            LOCKSROOT_OF_NO_LOCKS,
        ),
        {"from": A},
    )
    call_and_transact(token_network.functions.openChannel(A, B, TEST_SETTLE_TIMEOUT_MAX))
    channel_identifier = token_network.functions.getChannelIdentifier(A, B).call()
    (settle_block_number, _) = token_network.functions.getChannelInfo(
        channel_identifier, A, B
    ).call()

    assert settle_block_number == TEST_SETTLE_TIMEOUT_MAX



from typing import Callable

import pytest
from eth_tester.exceptions import TransactionFailed
from hexbytes import HexBytes
from web3 import Web3
from web3.contract import Contract
from web3.types import Wei

from raiden_contracts.constants import (
    TEST_SETTLE_TIMEOUT_MIN,
    ChannelEvent,
    ChannelState,
    MessageTypeId,
)
from raiden_contracts.tests.utils import (
    EMPTY_ADDITIONAL_HASH,
    EMPTY_BALANCE_HASH,
    EMPTY_SIGNATURE,
    NONEXISTENT_LOCKSROOT,
    ChannelValues,
    LockedAmounts,
    call_and_transact,
    fake_bytes,
)
from raiden_contracts.tests.utils.blockchain import mine_blocks
from raiden_contracts.utils.events import check_channel_closed


def test_close_nonexistent_channel(token_network: Contract, get_accounts: Callable) -> None:
    """Test getChannelInfo and closeChannel on a not-yet opened channel"""
    (A, B) = get_accounts(2)
    non_existent_channel_identifier = 1

    (settle_block_number, state) = token_network.functions.getChannelInfo(
        channel_identifier=non_existent_channel_identifier,
        participant1=A,
        participant2=B,
    ).call()
    assert state == ChannelState.NONEXISTENT
    assert settle_block_number == 0

    with pytest.raises(TransactionFailed, match="TN: channel not open"):
        token_network.functions.closeChannel(
            channel_identifier=non_existent_channel_identifier,
            closing_participant=A,
            non_closing_participant=B,
            balance_hash=EMPTY_BALANCE_HASH,
            nonce=0,
            additional_hash=EMPTY_ADDITIONAL_HASH,
            non_closing_signature=EMPTY_SIGNATURE,
            closing_signature=EMPTY_SIGNATURE,
        ).call({"from": A, "gas": Wei(81_000)})


def test_close_settled_channel_fail(
    web3: Web3,
    token_network: Contract,
    create_channel: Callable,
    channel_deposit: Callable,
    get_accounts: Callable,
    create_close_signature_for_no_balance_proof: Callable,
) -> None:
    """Test getChannelInfo and closeChannel on an already settled channel"""
    (A, B) = get_accounts(2)
    channel_identifier = create_channel(A, B, TEST_SETTLE_TIMEOUT_MIN)[0]
    channel_deposit(channel_identifier, A, 5, B)

    (_, state) = token_network.functions.getChannelInfo(channel_identifier, A, B).call()
    assert state == ChannelState.OPENED
    closing_sig = create_close_signature_for_no_balance_proof(A, channel_identifier)

    call_and_transact(
        token_network.functions.closeChannel(
            channel_identifier=channel_identifier,
            non_closing_participant=B,
            closing_participant=A,
            balance_hash=EMPTY_BALANCE_HASH,
            nonce=0,
            additional_hash=EMPTY_ADDITIONAL_HASH,
            non_closing_signature=EMPTY_SIGNATURE,
            closing_signature=closing_sig,
        ),
        {"from": A},
    )
    mine_blocks(web3, TEST_SETTLE_TIMEOUT_MIN + 1)
    call_and_transact(
        token_network.functions.settleChannel(
            channel_identifier=channel_identifier,
            participant1=A,
            participant1_transferred_amount=0,
            participant1_locked_amount=0,
            participant1_locksroot=NONEXISTENT_LOCKSROOT,
            participant2=B,
            participant2_transferred_amount=0,
            participant2_locked_amount=0,
            participant2_locksroot=NONEXISTENT_LOCKSROOT,
        ),
        {"from": A},
    )

    (settle_block_number, state) = token_network.functions.getChannelInfo(
        channel_identifier, A, B
    ).call()
    assert state == ChannelState.REMOVED
    assert settle_block_number == 0

    with pytest.raises(TransactionFailed, match=" TN: channel not open"):
        token_network.functions.closeChannel(
            channel_identifier=channel_identifier,
            non_closing_participant=B,
            closing_participant=A,
            balance_hash=EMPTY_BALANCE_HASH,
            nonce=0,
            additional_hash=EMPTY_ADDITIONAL_HASH,
            non_closing_signature=EMPTY_SIGNATURE,
            closing_signature=closing_sig,
        ).call({"from": A})


def test_close_wrong_signature(
    token_network: Contract,
    create_channel: Callable,
    channel_deposit: Callable,
    get_accounts: Callable,
    create_balance_proof: Callable,
    create_balance_proof_countersignature: Callable,
) -> None:
    """Closing a channel with a balance proof of the third party should fail"""
    (A, B, C) = get_accounts(3)
    deposit_A = 6
    transferred_amount = 5
    nonce = 3
    locksroot = fake_bytes(32, "03")

    channel_identifier = create_channel(A, B)[0]
    channel_deposit(channel_identifier, A, deposit_A, B)

    # Create balance proofs
    balance_proof = create_balance_proof(
        channel_identifier, C, transferred_amount, 0, nonce, locksroot
    )
    closing_signature_A = create_balance_proof_countersignature(
        participant=A,
        channel_identifier=channel_identifier,
        msg_type=MessageTypeId.BALANCE_PROOF,
        **balance_proof._asdict(),
    )

    with pytest.raises(TransactionFailed, match="TN/close: invalid non-closing sig"):
        token_network.functions.closeChannel(
            channel_identifier,
            B,
            A,
            *balance_proof._asdict().values(),
            closing_signature_A,
        ).call({"from": A})


def test_close_call_twice_fail(
    token_network: Contract,
    create_channel: Callable,
    channel_deposit: Callable,
    get_accounts: Callable,
    create_close_signature_for_no_balance_proof: Callable,
) -> None:
    """The second of two same closeChannel calls should fail"""
    (A, B) = get_accounts(2)
    channel_identifier = create_channel(A, B)[0]
    channel_deposit(channel_identifier, A, 5, B)

    closing_sig = create_close_signature_for_no_balance_proof(A, channel_identifier)

    call_and_transact(
        token_network.functions.closeChannel(
            channel_identifier=channel_identifier,
            non_closing_participant=B,
            closing_participant=A,
            balance_hash=EMPTY_BALANCE_HASH,
            nonce=0,
            additional_hash=EMPTY_ADDITIONAL_HASH,
            non_closing_signature=EMPTY_SIGNATURE,
            closing_signature=closing_sig,
        ),
        {"from": A},
    )

    with pytest.raises(TransactionFailed, match="TN: channel not open"):
        token_network.functions.closeChannel(
            channel_identifier=channel_identifier,
            non_closing_participant=B,
            closing_participant=A,
            balance_hash=EMPTY_BALANCE_HASH,
            nonce=0,
            additional_hash=EMPTY_ADDITIONAL_HASH,
            non_closing_signature=EMPTY_SIGNATURE,
            closing_signature=closing_sig,
        ).call({"from": A})


def test_close_different_sender(
    token_network: Contract,
    create_channel: Callable,
    channel_deposit: Callable,
    get_accounts: Callable,
    create_close_signature_for_no_balance_proof: Callable,
) -> None:
    """A closeChannel call from a different Ethereum address should succeed"""
    (A, B, C) = get_accounts(3)
    channel_identifier = create_channel(A, B)[0]
    channel_deposit(channel_identifier, A, 5, B)
    closing_sig = create_close_signature_for_no_balance_proof(A, channel_identifier)

    token_network.functions.closeChannel(
        channel_identifier=channel_identifier,
        non_closing_participant=B,
        closing_participant=A,
        balance_hash=EMPTY_BALANCE_HASH,
        nonce=0,
        additional_hash=EMPTY_ADDITIONAL_HASH,
        non_closing_signature=EMPTY_SIGNATURE,
        closing_signature=closing_sig,
    ).call({"from": C})


def test_close_nonce_zero(
    get_accounts: Callable,
    token_network: Contract,
    create_channel: Callable,
    create_balance_proof: Callable,
    create_balance_proof_countersignature: Callable,
    event_handler: Callable,
) -> None:
    """closeChannel with a balance proof with nonce zero should not change the channel state"""
    (A, B) = get_accounts(2)
    vals_B = ChannelValues(
        deposit=20,
        transferred=5,
        locksroot=fake_bytes(32, "03"),
        locked_amounts=LockedAmounts(claimable_locked=3, unclaimable_locked=4),
        nonce=0,
    )
    # Create channel and deposit
    channel_identifier = create_channel(A, B)[0]

    # Create balance proofs
    balance_proof_B = create_balance_proof(
        channel_identifier,
        B,
        vals_B.transferred,
        vals_B.locked_amounts.locked,
        vals_B.nonce,
        vals_B.locksroot,
    )
    close_sig_A = create_balance_proof_countersignature(
        participant=A,
        channel_identifier=channel_identifier,
        msg_type=MessageTypeId.BALANCE_PROOF,
        **balance_proof_B._asdict(),
    )

    (
        _,
        _,
        B_is_the_closer,
        B_balance_hash,
        B_nonce,
        _,
        _,
    ) = token_network.functions.getChannelParticipantInfo(channel_identifier, B, A).call()
    assert B_is_the_closer is False
    assert B_balance_hash == EMPTY_BALANCE_HASH
    assert B_nonce == 0

    ev_handler = event_handler(token_network)

    close_tx = call_and_transact(
        token_network.functions.closeChannel(
            channel_identifier, B, A, *balance_proof_B._asdict().values(), close_sig_A
        ),
        {"from": A},
    )

    ev_handler.add(
        close_tx,
        ChannelEvent.CLOSED,
        check_channel_closed(
            channel_identifier=channel_identifier,
            closing_participant=A,
            nonce=0,
            balance_hash=balance_proof_B.balance_hash,
        ),
    )
    ev_handler.check()

    # Even though we somehow provide valid values for the balance proof, they are not taken into
    # consideration if the nonce is 0.
    # The Raiden client enforces that the nonce is > 0 if off-chain transfers are made.
    (
        _,
        _,
        B_is_the_closer,
        B_balance_hash,
        B_nonce,
        _,
        _,
    ) = token_network.functions.getChannelParticipantInfo(channel_identifier, B, A).call()
    assert B_is_the_closer is False
    assert B_balance_hash == EMPTY_BALANCE_HASH
    assert B_nonce == 0


def test_close_first_argument_is_for_partner_transfer(
    token_network: Contract,
    create_channel: Callable,
    get_accounts: Callable,
    create_balance_proof: Callable,
    create_balance_proof_countersignature: Callable,
) -> None:
    """closeChannel fails on a self-submitted balance proof"""
    (A, B) = get_accounts(2)

    # Create channel
    channel_identifier = create_channel(A, B, settle_timeout=TEST_SETTLE_TIMEOUT_MIN)[0]

    # Create balance proofs
    balance_proof = create_balance_proof(channel_identifier, B)
    closing_sig_A = create_balance_proof_countersignature(
        participant=A,
        channel_identifier=channel_identifier,
        msg_type=MessageTypeId.BALANCE_PROOF,
        **balance_proof._asdict(),
    )
    closing_sig_B = create_balance_proof_countersignature(
        participant=B,
        channel_identifier=channel_identifier,
        msg_type=MessageTypeId.BALANCE_PROOF,
        **balance_proof._asdict(),
    )

    # closeChannel fails, if the provided balance proof is from the same participant who closes
    with pytest.raises(TransactionFailed, match="TN/close: invalid closing sig"):
        token_network.functions.closeChannel(
            channel_identifier, B, A, *balance_proof._asdict().values(), closing_sig_B
        ).call({"from": B})

    # Else, closeChannel works with this balance proof
    call_and_transact(
        token_network.functions.closeChannel(
            channel_identifier, B, A, *balance_proof._asdict().values(), closing_sig_A
        ),
        {"from": A},
    )


def test_close_first_participant_can_close(
    token_network: Contract,
    create_channel: Callable,
    get_accounts: Callable,
    get_block: Callable[[HexBytes], int],
    create_close_signature_for_no_balance_proof: Callable,
) -> None:
    """Simplest successful closeChannel by the first participant"""
    (A, B) = get_accounts(2)
    channel_identifier = create_channel(A, B)[0]

    closing_sig = create_close_signature_for_no_balance_proof(
        participant=A, channel_identifier=channel_identifier
    )
    close_tx = call_and_transact(
        token_network.functions.closeChannel(
            channel_identifier=channel_identifier,
            non_closing_participant=B,
            closing_participant=A,
            balance_hash=EMPTY_BALANCE_HASH,
            nonce=0,
            additional_hash=EMPTY_ADDITIONAL_HASH,
            non_closing_signature=EMPTY_SIGNATURE,
            closing_signature=closing_sig,
        ),
        {"from": A},
    )

    (settle_block_number, state) = token_network.functions.getChannelInfo(
        channel_identifier, B, A
    ).call()
    assert settle_block_number == TEST_SETTLE_TIMEOUT_MIN + get_block(close_tx)
    assert state == ChannelState.CLOSED

    (
        _,
        _,
        A_is_the_closer,
        A_balance_hash,
        A_nonce,
        _,
        _,
    ) = token_network.functions.getChannelParticipantInfo(channel_identifier, A, B).call()
    assert A_is_the_closer is True
    assert A_balance_hash == EMPTY_BALANCE_HASH
    assert A_nonce == 0

    (
        _,
        _,
        B_is_the_closer,
        B_balance_hash,
        B_nonce,
        _,
        _,
    ) = token_network.functions.getChannelParticipantInfo(channel_identifier, B, A).call()
    assert B_is_the_closer is False
    assert B_balance_hash == EMPTY_BALANCE_HASH
    assert B_nonce == 0


def test_close_second_participant_can_close(
    token_network: Contract,
    create_channel: Callable,
    get_accounts: Callable,
    create_close_signature_for_no_balance_proof: Callable,
) -> None:
    """Simplest successful closeChannel by the second participant"""
    (A, B) = get_accounts(2)
    channel_identifier = create_channel(A, B)[0]
    closing_sig = create_close_signature_for_no_balance_proof(B, channel_identifier)

    call_and_transact(
        token_network.functions.closeChannel(
            channel_identifier=channel_identifier,
            non_closing_participant=A,
            closing_participant=B,
            balance_hash=EMPTY_BALANCE_HASH,
            nonce=0,
            additional_hash=EMPTY_ADDITIONAL_HASH,
            non_closing_signature=EMPTY_SIGNATURE,
            closing_signature=closing_sig,
        ),
        {"from": B},
    )


def test_close_channel_state(
    web3: Web3,
    custom_token: Contract,
    token_network: Contract,
    create_channel: Callable,
    channel_deposit: Callable,
    get_accounts: Callable,
    get_block: Callable,
    create_balance_proof: Callable,
    txn_cost: Callable,
    create_balance_proof_countersignature: Callable,
) -> None:
    """Observe the effect of a successful closeChannel

    This test compares the state of the channel and the balances of Ethereum
    accounts before/after a successful closeChannel call.
    """
    (A, B) = get_accounts(2)
    settle_timeout = TEST_SETTLE_TIMEOUT_MIN
    vals_B = ChannelValues(
        deposit=20,
        transferred=5,
        locksroot=fake_bytes(32, "03"),
        locked_amounts=LockedAmounts(claimable_locked=3, unclaimable_locked=4),
        nonce=3,
    )

    # Create channel and deposit
    channel_identifier = create_channel(A, B, settle_timeout)[0]
    channel_deposit(channel_identifier, B, vals_B.deposit, A)

    # Check the state of the openned channel
    (settle_block_number, state) = token_network.functions.getChannelInfo(
        channel_identifier, A, B
    ).call()
    assert settle_block_number == settle_timeout
    assert state == ChannelState.OPENED

    (
        _,
        _,
        A_is_the_closer,
        A_balance_hash,
        A_nonce,
        _,
        _,
    ) = token_network.functions.getChannelParticipantInfo(channel_identifier, A, B).call()
    assert A_is_the_closer is False
    assert A_balance_hash == EMPTY_BALANCE_HASH
    assert A_nonce == 0
    (
        _,
        _,
        B_is_the_closer,
        B_balance_hash,
        B_nonce,
        _,
        _,
    ) = token_network.functions.getChannelParticipantInfo(channel_identifier, B, A).call()
    assert B_is_the_closer is False
    assert B_balance_hash == EMPTY_BALANCE_HASH
    assert B_nonce == 0

    pre_eth_balance_A = web3.eth.get_balance(A)
    pre_eth_balance_B = web3.eth.get_balance(B)
    pre_eth_balance_contract = web3.eth.get_balance(token_network.address)
    pre_balance_A = custom_token.functions.balanceOf(A).call()
    pre_balance_B = custom_token.functions.balanceOf(B).call()
    pre_balance_contract = custom_token.functions.balanceOf(token_network.address).call()

    # Create a balance proof
    balance_proof_B = create_balance_proof(
        channel_identifier,
        B,
        vals_B.transferred,
        vals_B.locked_amounts.locked,
        vals_B.nonce,
        vals_B.locksroot,
    )
    closing_sig_A = create_balance_proof_countersignature(
        participant=A,
        channel_identifier=channel_identifier,
        msg_type=MessageTypeId.BALANCE_PROOF,
        **balance_proof_B._asdict(),
    )

    txn_hash = call_and_transact(
        token_network.functions.closeChannel(
            channel_identifier, B, A, *balance_proof_B._asdict().values(), closing_sig_A
        ),
        {"from": A},
    )

    # Test that no balances have changed.
    # There are no transfers to be made in closeChannel.
    assert web3.eth.get_balance(A) == pre_eth_balance_A - txn_cost(txn_hash)
    assert web3.eth.get_balance(B) == pre_eth_balance_B
    assert web3.eth.get_balance(token_network.address) == pre_eth_balance_contract
    assert custom_token.functions.balanceOf(A).call() == pre_balance_A
    assert custom_token.functions.balanceOf(B).call() == pre_balance_B
    assert custom_token.functions.balanceOf(token_network.address).call() == pre_balance_contract

    (settle_block_number, state) = token_network.functions.getChannelInfo(
        channel_identifier, A, B
    ).call()
    assert settle_block_number == settle_timeout + get_block(txn_hash)
    assert state == ChannelState.CLOSED

    (
        _,
        _,
        A_is_the_closer,
        A_balance_hash,
        A_nonce,
        _,
        _,
    ) = token_network.functions.getChannelParticipantInfo(channel_identifier, A, B).call()
    assert A_is_the_closer is True
    assert A_balance_hash == EMPTY_BALANCE_HASH
    assert A_nonce == 0

    (
        _,
        _,
        B_is_the_closer,
        B_balance_hash,
        B_nonce,
        _,
        _,
    ) = token_network.functions.getChannelParticipantInfo(channel_identifier, B, A).call()
    assert B_is_the_closer is False
    assert B_balance_hash == balance_proof_B.balance_hash
    assert B_nonce == vals_B.nonce


def test_close_channel_event_no_offchain_transfers(
    get_accounts: Callable,
    token_network: Contract,
    create_channel: Callable,
    event_handler: Callable,
    create_close_signature_for_no_balance_proof: Callable,
) -> None:
    """closeChannel succeeds and emits an event even with nonce 0 and no balance proofs"""
    ev_handler = event_handler(token_network)
    (A, B) = get_accounts(2)

    channel_identifier = create_channel(A, B)[0]
    closing_sig = create_close_signature_for_no_balance_proof(A, channel_identifier)

    # No off-chain transfers have occured
    # There is no signature data here, because it was never provided to A
    txn_hash = call_and_transact(
        token_network.functions.closeChannel(
            channel_identifier,
            B,
            A,
            EMPTY_BALANCE_HASH,
            0,
            EMPTY_ADDITIONAL_HASH,
            EMPTY_SIGNATURE,
            closing_sig,
        ),
        {"from": A},
    )

    ev_handler.add(
        txn_hash,
        ChannelEvent.CLOSED,
        check_channel_closed(
            channel_identifier=channel_identifier,
            closing_participant=A,
            nonce=0,
            balance_hash=EMPTY_BALANCE_HASH,
        ),
    )
    ev_handler.check()


def test_close_replay_reopened_channel(
    web3: Web3,
    get_accounts: Callable,
    token_network: Contract,
    create_channel: Callable,
    channel_deposit: Callable,
    create_balance_proof: Callable,
    create_balance_proof_countersignature: Callable,
) -> None:
    """The same balance proof cannot close another channel between the same participants"""
    (A, B) = get_accounts(2)
    nonce = 3
    values_A = ChannelValues(deposit=10, transferred=0)
    values_B = ChannelValues(deposit=20, transferred=15)
    channel_identifier1 = create_channel(A, B)[0]
    channel_deposit(channel_identifier1, B, values_B.deposit, A)

    balance_proof_B = create_balance_proof(
        channel_identifier1,
        B,
        values_B.transferred,
        values_B.locked_amounts.locked,
        nonce,
        values_B.locksroot,
    )
    closing_sig_A = create_balance_proof_countersignature(
        participant=A,
        channel_identifier=channel_identifier1,
        msg_type=MessageTypeId.BALANCE_PROOF,
        **balance_proof_B._asdict(),
    )
    call_and_transact(
        token_network.functions.closeChannel(
            channel_identifier1,
            B,
            A,
            *balance_proof_B._asdict().values(),
            closing_sig_A,
        ),
        {"from": A},
    )
    mine_blocks(web3, TEST_SETTLE_TIMEOUT_MIN + 1)
    call_and_transact(
        token_network.functions.settleChannel(
            channel_identifier=channel_identifier1,
            participant1=A,
            participant1_transferred_amount=values_A.transferred,
            participant1_locked_amount=values_A.locked_amounts.locked,
            participant1_locksroot=values_A.locksroot,
            participant2=B,
            participant2_transferred_amount=values_B.transferred,
            participant2_locked_amount=values_B.locked_amounts.locked,
            participant2_locksroot=values_B.locksroot,
        ),
        {"from": A},
    )

    # Reopen the channel and make sure we cannot use the old balance proof
    channel_identifier2 = create_channel(A, B)[0]
    channel_deposit(channel_identifier2, B, values_B.deposit, A)

    assert channel_identifier1 != channel_identifier2
    with pytest.raises(TransactionFailed, match="TN/close: invalid closing sig"):
        token_network.functions.closeChannel(
            channel_identifier2,
            B,
            A,
            *balance_proof_B._asdict().values(),
            closing_sig_A,
        ).call({"from": A})

    # Balance proof with correct channel_identifier must work
    balance_proof_B2 = create_balance_proof(
        channel_identifier2,
        B,
        values_B.transferred,
        values_B.locked_amounts.locked,
        nonce,
        values_B.locksroot,
    )
    closing_sig_A2 = create_balance_proof_countersignature(
        participant=A,
        channel_identifier=channel_identifier2,
        msg_type=MessageTypeId.BALANCE_PROOF,
        **balance_proof_B2._asdict(),
    )
    call_and_transact(
        token_network.functions.closeChannel(
            channel_identifier2,
            B,
            A,
            *balance_proof_B2._asdict().values(),
            closing_sig_A2,
        ),
        {"from": A},
    )


def test_close_channel_event(
    get_accounts: Callable,
    token_network: Contract,
    create_channel: Callable,
    channel_deposit: Callable,
    create_balance_proof: Callable,
    event_handler: Callable,
    create_balance_proof_countersignature: Callable,
) -> None:
    """A successful closeChannel call produces a CLOSED event"""
    ev_handler = event_handler(token_network)
    (A, B) = get_accounts(2)
    deposit_A = 10

    channel_identifier = create_channel(A, B)[0]
    channel_deposit(channel_identifier, A, deposit_A, B)
    balance_proof = create_balance_proof(
        channel_identifier, B, transferred_amount=5, locked_amount=0, nonce=3
    )
    close_sig = create_balance_proof_countersignature(
        participant=A,
        channel_identifier=channel_identifier,
        msg_type=MessageTypeId.BALANCE_PROOF,
        **balance_proof._asdict(),
    )

    txn_hash = call_and_transact(
        token_network.functions.closeChannel(
            channel_identifier, B, A, *balance_proof._asdict().values(), close_sig
        ),
        {"from": A},
    )

    ev_handler.add(
        txn_hash,
        ChannelEvent.CLOSED,
        check_channel_closed(
            channel_identifier=channel_identifier,
            closing_participant=A,
            nonce=3,
            balance_hash=balance_proof.balance_hash,
        ),
    )
    ev_handler.check()



from itertools import chain, product

import pytest
from eth_tester.constants import UINT256_MAX, UINT256_MIN
from web3.contract import Contract
from web3.exceptions import ValidationError

from raiden_contracts.contract_source_manager import check_runtime_codesize


def test_min_uses_usigned(token_network: Contract) -> None:
    """Min cannot be called with negative values"""
    INVALID_VALUES = [-UINT256_MAX, -1]
    VALID_VALUES = [UINT256_MIN, UINT256_MAX, UINT256_MAX]

    all_invalid = chain(
        product(VALID_VALUES, INVALID_VALUES), product(INVALID_VALUES, VALID_VALUES)
    )

    for a, b in all_invalid:
        with pytest.raises(ValidationError):
            token_network.functions.min(a, b).call()


def test_min(token_network: Contract) -> None:
    """Min works like Python's min"""
    VALUES = [UINT256_MIN, 1, UINT256_MAX, UINT256_MAX]
    for a, b in product(VALUES, VALUES):
        assert token_network.functions.min(a, b).call() == min(a, b)


def test_too_big_runtime_code() -> None:
    compilation = {"TestContract": {"bin-runtime": "33" * 0x6001}}
    with pytest.raises(RuntimeError):
        check_runtime_codesize(compilation)



#!/usr/bin/env python3
try:
    from setuptools import setup, find_packages
except ImportError:
    from distutils.core import setup

import os
from typing import List

from setuptools import Command
from setuptools.command.build_py import build_py

DESCRIPTION = "Raiden contracts library and utilities"
VERSION = "0.39.0"


def read_requirements(path: str) -> List[str]:
    assert os.path.isfile(path)
    with open(path) as requirements:
        return requirements.read().split()


def _get_single_requirement(requirements: List[str], package: str) -> List[str]:
    return [req for req in requirements if req.startswith(package)]


class BuildPyCommand(build_py):
    def run(self) -> None:
        try:
            self.run_command("verify_contracts")
        except SystemExit:
            pass
        build_py.run(self)


class VerifyContracts(Command):
    description = "Verify that the compiled contracts have the correct source code checksum"
    user_options: List = []

    def initialize_options(self) -> None:
        pass

    def finalize_options(self) -> None:
        pass

    def run(self) -> None:  # pylint: disable=no-self-use
        from raiden_contracts.contract_manager import contracts_precompiled_path
        from raiden_contracts.contract_source_manager import (
            ContractSourceManager,
            contracts_source_path,
        )

        manager = ContractSourceManager(contracts_source_path(contracts_version=None))
        manager.verify_precompiled_checksums(contracts_precompiled_path())


class CompileContracts(Command):
    description = "Compile contracts and add ABI and checksums to a json file"
    user_options: List = []

    def initialize_options(self) -> None:
        pass

    def finalize_options(self) -> None:
        pass

    def run(self) -> None:  # pylint: disable=no-self-use
        from raiden_contracts.contract_manager import contracts_precompiled_path
        from raiden_contracts.contract_source_manager import (
            ContractSourceManager,
            contracts_source_path,
        )

        contract_manager = ContractSourceManager(contracts_source_path(contracts_version=None))
        contract_manager.compile_contracts(contracts_precompiled_path())


requirements = read_requirements("requirements.txt")

config = {
    "version": VERSION,
    "scripts": [],
    "name": "raiden-contracts",
    "author": "Brainbot Labs Est.",
    "author_email": "contact@brainbot.li",
    "description": DESCRIPTION,
    "url": "https://github.com/raiden-network/raiden-contracts/",
    "license": "MIT",
    "keywords": "raiden ethereum blockchain",
    "install_requires": requirements,
    "setup_requires": requirements,
    "packages": find_packages(),
    "include_package_data": True,
    "classifiers": [
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    "entry_points": {"console_scripts": ["deploy = raiden_contracts.deploy.__main__:main"]},
    "cmdclass": {
        "compile_contracts": CompileContracts,
        "verify_contracts": VerifyContracts,
        "build_py": BuildPyCommand,
    },
    "zip_safe": False,
    "package_data": {"raiden_contracts": ["py.typed"]},
}

setup(**config)



from eth_abi import encode_single
from eth_typing.evm import HexAddress
from web3 import Web3
from web3.types import Nonce

from raiden_contracts.constants import MessageTypeId
from raiden_contracts.utils.type_aliases import (
    AdditionalHash,
    BalanceHash,
    BlockExpiration,
    ChainID,
    ChannelID,
    Locksroot,
    PrivateKey,
    Signature,
    TokenAmount,
)

from .signature import sign


def hash_balance_data(
    transferred_amount: TokenAmount, locked_amount: TokenAmount, locksroot: Locksroot
) -> BalanceHash:
    # pylint: disable=E1120
    return Web3.solidityKeccak(
        abi_types=["uint256", "uint256", "bytes32"],
        values=[transferred_amount, locked_amount, locksroot],
    )


def eth_sign_hash_message(encoded_message: bytes) -> bytes:
    signature_prefix = "\x19Ethereum Signed Message:\n"
    return Web3.keccak(
        Web3.toBytes(text=signature_prefix)
        + Web3.toBytes(text=str(len(encoded_message)))
        + encoded_message
    )


def pack_balance_proof(
    token_network_address: HexAddress,
    chain_identifier: ChainID,
    channel_identifier: ChannelID,
    balance_hash: BalanceHash,
    nonce: Nonce,
    additional_hash: AdditionalHash,
    msg_type: MessageTypeId,
) -> bytes:
    return (
        Web3.toBytes(hexstr=token_network_address)
        + encode_single("uint256", chain_identifier)
        + encode_single("uint256", msg_type)
        + encode_single("uint256", channel_identifier)
        + balance_hash
        + encode_single("uint256", nonce)
        + additional_hash
    )


def pack_balance_proof_message(
    token_network_address: HexAddress,
    chain_identifier: ChainID,
    channel_identifier: ChannelID,
    msg_type: MessageTypeId,
    balance_hash: BalanceHash,
    nonce: Nonce,
    additional_hash: AdditionalHash,
    closing_signature: Signature,
) -> bytes:
    return (
        pack_balance_proof(
            token_network_address=token_network_address,
            chain_identifier=chain_identifier,
            channel_identifier=channel_identifier,
            balance_hash=balance_hash,
            nonce=nonce,
            additional_hash=additional_hash,
            msg_type=msg_type,
        )
        + closing_signature
    )


def pack_cooperative_settle_message(
    token_network_address: HexAddress,
    chain_identifier: ChainID,
    channel_identifier: ChannelID,
    participant1_address: HexAddress,
    participant1_balance: TokenAmount,
    participant2_address: HexAddress,
    participant2_balance: TokenAmount,
) -> bytes:
    return (
        Web3.toBytes(hexstr=token_network_address)
        + encode_single("uint256", chain_identifier)
        + encode_single("uint256", MessageTypeId.COOPERATIVE_SETTLE)
        + encode_single("uint256", channel_identifier)
        + Web3.toBytes(hexstr=participant1_address)
        + encode_single("uint256", participant1_balance)
        + Web3.toBytes(hexstr=participant2_address)
        + encode_single("uint256", participant2_balance)
    )


def pack_withdraw_message(
    token_network_address: HexAddress,
    chain_identifier: ChainID,
    channel_identifier: ChannelID,
    participant: HexAddress,
    amount_to_withdraw: TokenAmount,
    expiration_block: BlockExpiration,
) -> bytes:
    return (
        Web3.toBytes(hexstr=token_network_address)
        + encode_single("uint256", chain_identifier)
        + encode_single("uint256", MessageTypeId.WITHDRAW)
        + encode_single("uint256", channel_identifier)
        + Web3.toBytes(hexstr=participant)
        + encode_single("uint256", amount_to_withdraw)
        + encode_single("uint256", expiration_block)
    )


def pack_reward_proof(
    monitoring_service_contract_address: HexAddress,
    chain_id: ChainID,
    token_network_address: HexAddress,
    non_closing_participant: HexAddress,
    non_closing_signature: Signature,
    reward_amount: TokenAmount,
) -> bytes:
    return (
        Web3.toBytes(hexstr=monitoring_service_contract_address)
        + encode_single("uint256", chain_id)
        + encode_single("uint256", MessageTypeId.MSReward)
        + Web3.toBytes(hexstr=token_network_address)
        + Web3.toBytes(hexstr=non_closing_participant)
        + non_closing_signature
        + encode_single("uint256", reward_amount)
    )


def sign_balance_proof(
    privatekey: PrivateKey,
    token_network_address: HexAddress,
    chain_identifier: ChainID,
    channel_identifier: ChannelID,
    msg_type: MessageTypeId,
    balance_hash: BalanceHash,
    nonce: Nonce,
    additional_hash: AdditionalHash,
    v: int = 27,
) -> bytes:
    message_hash = eth_sign_hash_message(
        pack_balance_proof(
            token_network_address=token_network_address,
            chain_identifier=chain_identifier,
            channel_identifier=channel_identifier,
            balance_hash=balance_hash,
            nonce=nonce,
            additional_hash=additional_hash,
            msg_type=msg_type,
        )
    )

    return sign(privkey=privatekey, msg_hash=message_hash, v=v)


def sign_balance_proof_message(
    privatekey: PrivateKey,
    token_network_address: HexAddress,
    chain_identifier: ChainID,
    channel_identifier: ChannelID,
    msg_type: MessageTypeId,
    balance_hash: BalanceHash,
    nonce: Nonce,
    additional_hash: AdditionalHash,
    closing_signature: Signature,
    v: int = 27,
) -> bytes:
    message_hash = eth_sign_hash_message(
        pack_balance_proof_message(
            token_network_address=token_network_address,
            chain_identifier=chain_identifier,
            channel_identifier=channel_identifier,
            msg_type=msg_type,
            balance_hash=balance_hash,
            nonce=nonce,
            additional_hash=additional_hash,
            closing_signature=closing_signature,
        )
    )

    return sign(privkey=privatekey, msg_hash=message_hash, v=v)


def sign_withdraw_message(
    privatekey: PrivateKey,
    token_network_address: HexAddress,
    chain_identifier: ChainID,
    channel_identifier: ChannelID,
    participant: HexAddress,
    amount_to_withdraw: TokenAmount,
    expiration_block: BlockExpiration,
    v: int = 27,
) -> bytes:
    message_hash = eth_sign_hash_message(
        pack_withdraw_message(
            token_network_address=token_network_address,
            chain_identifier=chain_identifier,
            channel_identifier=channel_identifier,
            participant=participant,
            amount_to_withdraw=amount_to_withdraw,
            expiration_block=expiration_block,
        )
    )

    return sign(privkey=privatekey, msg_hash=message_hash, v=v)


def sign_reward_proof(
    privatekey: PrivateKey,
    monitoring_service_contract_address: HexAddress,
    chain_id: ChainID,
    token_network_address: HexAddress,
    non_closing_participant: HexAddress,
    non_closing_signature: Signature,
    reward_amount: TokenAmount,
    v: int = 27,
) -> bytes:
    packed_data = pack_reward_proof(
        monitoring_service_contract_address=monitoring_service_contract_address,
        chain_id=chain_id,
        token_network_address=token_network_address,
        non_closing_participant=non_closing_participant,
        non_closing_signature=non_closing_signature,
        reward_amount=reward_amount,
    )
    message_hash = eth_sign_hash_message(packed_data)

    return sign(privkey=privatekey, msg_hash=message_hash, v=v)


def sign_one_to_n_iou(
    privatekey: PrivateKey,
    sender: HexAddress,
    receiver: HexAddress,
    amount: TokenAmount,
    expiration_block: BlockExpiration,
    one_to_n_address: HexAddress,
    chain_id: ChainID,
    v: int = 27,
) -> bytes:
    iou_hash = eth_sign_hash_message(
        Web3.toBytes(hexstr=one_to_n_address)
        + encode_single("uint256", chain_id)
        + encode_single("uint256", MessageTypeId.IOU)
        + Web3.toBytes(hexstr=sender)
        + Web3.toBytes(hexstr=receiver)
        + encode_single("uint256", amount)
        + encode_single("uint256", expiration_block)
    )
    return sign(privkey=privatekey, msg_hash=iou_hash, v=v)



import json
from typing import Any, List, Optional, Tuple

from eth_typing.evm import HexAddress
from eth_utils import to_checksum_address
from web3 import Web3
from web3.contract import Contract

from raiden_contracts.constants import (
    CONTRACT_MONITORING_SERVICE,
    CONTRACT_ONE_TO_N,
    CONTRACT_SECRET_REGISTRY,
    CONTRACT_SERVICE_REGISTRY,
    CONTRACT_TOKEN_NETWORK_REGISTRY,
    CONTRACT_USER_DEPOSIT,
    DeploymentModule,
)
from raiden_contracts.contract_manager import (
    ContractManager,
    DeployedContracts,
    contracts_deployed_path,
    contracts_precompiled_path,
    get_contracts_deployment_info,
)
from raiden_contracts.utils.type_aliases import ChainID


class ContractVerifier:
    def __init__(self, web3: Web3, contracts_version: Optional[str] = None):
        self.web3 = web3
        self.contracts_version = contracts_version
        self.precompiled_path = contracts_precompiled_path(self.contracts_version)
        self.contract_manager = ContractManager(self.precompiled_path)

    def verify_deployed_contracts_in_filesystem(self) -> None:
        chain_id = ChainID(self.web3.eth.chain_id)

        deployment_data = get_contracts_deployment_info(
            chain_id=chain_id,
            version=self.contract_manager.contracts_version,
            module=DeploymentModule.RAIDEN,
        )
        deployment_file_path = contracts_deployed_path(
            chain_id=chain_id, version=self.contract_manager.contracts_version
        )
        if deployment_data is None:
            raise RuntimeError(f"Deployment data cannot be found at {deployment_file_path}")

        if self.verify_deployment_data(deployment_data):
            print(
                f"Deployment info from {deployment_file_path} has been verified "
                "and it is CORRECT."
            )

    def verify_deployed_service_contracts_in_filesystem(
        self,
        token_address: HexAddress,
        user_deposit_whole_balance_limit: int,
        token_network_registry_address: HexAddress,
    ) -> None:
        chain_id = ChainID(self.web3.eth.chain_id)

        deployment_data = get_contracts_deployment_info(
            chain_id=chain_id,
            version=self.contract_manager.contracts_version,
            module=DeploymentModule.SERVICES,
        )
        deployment_file_path = contracts_deployed_path(
            chain_id=chain_id, version=self.contract_manager.contracts_version, services=True
        )
        if deployment_data is None:
            raise RuntimeError(f"Deployment data cannot be found at {deployment_file_path}")

        if self.verify_service_contracts_deployment_data(
            token_address=token_address,
            user_deposit_whole_balance_limit=user_deposit_whole_balance_limit,
            deployed_contracts_info=deployment_data,
            token_network_registry_address=token_network_registry_address,
        ):
            print(
                f"Deployment info from {deployment_file_path} has been verified "
                "and it is CORRECT."
            )

    def store_and_verify_deployment_info_raiden(
        self, deployed_contracts_info: DeployedContracts
    ) -> None:
        self._store_deployment_info(deployment_info=deployed_contracts_info, services=False)
        self.verify_deployed_contracts_in_filesystem()

    def store_and_verify_deployment_info_services(
        self,
        deployed_contracts_info: DeployedContracts,
        token_address: HexAddress,
        user_deposit_whole_balance_limit: int,
        token_network_registry_address: HexAddress,
    ) -> None:
        self._store_deployment_info(services=True, deployment_info=deployed_contracts_info)
        self.verify_deployed_service_contracts_in_filesystem(
            token_address=token_address,
            user_deposit_whole_balance_limit=user_deposit_whole_balance_limit,
            token_network_registry_address=token_network_registry_address,
        )

    def _store_deployment_info(self, services: bool, deployment_info: DeployedContracts) -> None:
        deployment_file_path = contracts_deployed_path(
            chain_id=ChainID(self.web3.eth.chain_id),
            version=self.contracts_version,
            services=services,
        )
        with deployment_file_path.open(mode="w") as target_file:
            target_file.write(json.dumps(deployment_info, indent=2))

        print(
            f'Deployment information for chain id = {deployment_info["chain_id"]} '
            f" has been updated at {deployment_file_path}."
        )

    def verify_deployment_data(self, deployment_data: DeployedContracts) -> bool:
        chain_id = self.web3.eth.chain_id

        if self.contract_manager.contracts_version != deployment_data["contracts_version"]:
            raise RuntimeError("Version string mismatch.")
        if chain_id != deployment_data["chain_id"]:
            raise RuntimeError("chain id mismatch.")

        secret_registry, _ = self._verify_deployed_contract(
            deployment_data=deployment_data, contract_name=CONTRACT_SECRET_REGISTRY
        )

        token_network_registry, constructor_arguments = self._verify_deployed_contract(
            deployment_data=deployment_data, contract_name=CONTRACT_TOKEN_NETWORK_REGISTRY
        )

        # We need to also check the constructor parameters against the chain
        if (
            to_checksum_address(token_network_registry.functions.secret_registry_address().call())
            != secret_registry.address
        ):
            raise RuntimeError("secret_registry_address onchain has an unexpected value.")
        if len(constructor_arguments) != 5:
            raise RuntimeError(
                "TokenNetworkRegistry received a wrong number of constructor arguments."
            )
        if secret_registry.address != constructor_arguments[0]:
            raise RuntimeError(
                "TokenNetworkRegistry's constructor received a different SecretRegistry address."
            )
        if token_network_registry.functions.chain_id().call() != constructor_arguments[1]:
            raise RuntimeError("TokenNetwork remembers a wrong chain_id.")
        assert (
            token_network_registry.functions.settlement_timeout_min().call()
            == constructor_arguments[2]
        )
        assert (
            token_network_registry.functions.settlement_timeout_max().call()
            == constructor_arguments[3]
        )

        return True

    def _verify_deployed_contract(
        self, deployment_data: DeployedContracts, contract_name: str
    ) -> Tuple[Contract, List[Any]]:
        """Verify deployment info against the chain

        Verifies:
        - the runtime bytecode - precompiled data against the chain
        - information stored in deployment_*.json against the chain,
        except for the constructor arguments, which have to be checked
        separately.

        Returns: (onchain_instance, constructor_arguments)
        """
        contract_instance = self.contract_instance_from_deployment_data(
            deployment_data, contract_name
        )
        contracts = deployment_data["contracts"]

        # Check blockchain transaction hash & block information
        receipt = self.web3.eth.get_transaction_receipt(
            contracts[contract_name]["transaction_hash"]
        )
        if receipt["blockNumber"] != contracts[contract_name]["block_number"]:
            raise RuntimeError(
                f'We have block_number {contracts[contract_name]["block_number"]} in the '
                f'deployment info, but {receipt["blockNumber"]} in the transaction receipt '
                "from web3."
            )
        if receipt["gasUsed"] != contracts[contract_name]["gas_cost"]:
            raise RuntimeError(
                f'We have gasUsed {contracts[contract_name]["gas_cost"]} in the deployment info, '
                f'but {receipt["gasUsed"]} in the transaction receipt from web3.'
            )
        if receipt["contractAddress"] != contracts[contract_name]["address"]:
            raise RuntimeError(
                f'We have contractAddress {contracts[contract_name]["address"]} in the deployment'
                f' info but {receipt["contractAddress"]} in the transaction receipt from web3.'
            )

        # Check that the deployed bytecode matches the precompiled data
        blockchain_bytecode = self.web3.eth.get_code(contract_instance.address).hex()
        compiled_bytecode = self.contract_manager.get_runtime_hexcode(contract_name)

        if blockchain_bytecode == compiled_bytecode:
            print(
                f"{contract_name} at {contract_instance.address} "
                f"matches the compiled data from contracts.json"
            )
        else:
            raise RuntimeError(f"{contract_name} at {contract_instance.address} has wrong code")

        return contract_instance, contracts[contract_name]["constructor_arguments"]

    def contract_instance_from_deployment_data(
        self, deployment_data: DeployedContracts, contract_name: str
    ) -> Contract:
        contracts = deployment_data["contracts"]
        contract_address = contracts[contract_name]["address"]
        contract_instance = self.web3.eth.contract(
            abi=self.contract_manager.get_contract_abi(contract_name), address=contract_address
        )
        return contract_instance

    def verify_service_contracts_deployment_data(
        self,
        token_address: HexAddress,
        user_deposit_whole_balance_limit: int,
        token_network_registry_address: HexAddress,
        deployed_contracts_info: DeployedContracts,
    ) -> bool:
        chain_id = self.web3.eth.chain_id
        assert deployed_contracts_info is not None

        if self.contract_manager.contracts_version != deployed_contracts_info["contracts_version"]:
            raise RuntimeError("Version string mismatch")
        if chain_id != deployed_contracts_info["chain_id"]:
            raise RuntimeError("chain_id mismatch")

        (
            service_registry,
            service_registry_constructor_arguments,
        ) = self._verify_deployed_contract(
            deployment_data=deployed_contracts_info, contract_name=CONTRACT_SERVICE_REGISTRY
        )
        (user_deposit, user_deposit_constructor_arguments,) = self._verify_deployed_contract(
            deployment_data=deployed_contracts_info, contract_name=CONTRACT_USER_DEPOSIT
        )
        one_to_n, one_to_n_constructor_arguments = self._verify_deployed_contract(
            deployment_data=deployed_contracts_info, contract_name=CONTRACT_ONE_TO_N
        )
        monitoring_service, ms_constructor_arguments = self._verify_deployed_contract(
            deployed_contracts_info, CONTRACT_MONITORING_SERVICE
        )
        _verify_service_registry_deployment(
            service_registry=service_registry,
            constructor_arguments=service_registry_constructor_arguments,
            token_address=token_address,
        )
        _verify_user_deposit_deployment(
            user_deposit=user_deposit,
            constructor_arguments=user_deposit_constructor_arguments,
            token_address=token_address,
            user_deposit_whole_balance_limit=user_deposit_whole_balance_limit,
            one_to_n_address=one_to_n.address,
            monitoring_service_address=monitoring_service.address,
        )
        _verify_monitoring_service_deployment(
            monitoring_service=monitoring_service,
            constructor_arguments=ms_constructor_arguments,
            token_address=token_address,
            service_registry_address=service_registry.address,
            user_deposit_address=user_deposit.address,
            token_network_registry_address=token_network_registry_address,
        )
        _verify_one_to_n_deployment(
            one_to_n=one_to_n,
            constructor_arguments=one_to_n_constructor_arguments,
            user_deposit_address=user_deposit.address,
            chain_id=chain_id,
            service_registry_address=service_registry.address,
        )
        return True


def _verify_user_deposit_deployment(
    user_deposit: Contract,
    constructor_arguments: List,
    token_address: HexAddress,
    user_deposit_whole_balance_limit: int,
    one_to_n_address: HexAddress,
    monitoring_service_address: HexAddress,
) -> None:
    """Check an onchain deployment of UserDeposit and constructor arguments at deployment time"""
    if len(constructor_arguments) != 2:
        raise RuntimeError("UserDeposit has a wrong number of constructor arguments.")
    if token_address != constructor_arguments[0]:
        raise RuntimeError("UserDeposit received a wrong token address during construction.")
    if to_checksum_address(user_deposit.functions.token().call()) != token_address:
        raise RuntimeError("UserDeposit has a wrong token address onchain.")
    if user_deposit.functions.whole_balance_limit().call() != user_deposit_whole_balance_limit:
        raise RuntimeError("UserDeposit has a wrong whole_balance_limit onchain")
    if user_deposit_whole_balance_limit != constructor_arguments[1]:
        raise RuntimeError("UserDeposit received a wrong whole_balance_limit during construction.")
    if to_checksum_address(user_deposit.functions.one_to_n_address().call()) != one_to_n_address:
        raise RuntimeError("UserDeposit has a wrong OneToN address onchain.")
    onchain_msc_address = to_checksum_address(user_deposit.functions.msc_address().call())
    if onchain_msc_address != monitoring_service_address:
        raise RuntimeError(
            f"MSC address found onchain: {onchain_msc_address}, "
            f"expected: {monitoring_service_address}"
        )


def _verify_monitoring_service_deployment(
    monitoring_service: Contract,
    constructor_arguments: List,
    token_address: HexAddress,
    service_registry_address: HexAddress,
    user_deposit_address: HexAddress,
    token_network_registry_address: HexAddress,
) -> None:
    """Check an onchain deployment of MonitoringService and constructor arguments"""
    if len(constructor_arguments) != 4:
        raise RuntimeError("MonitoringService has a wrong number of constructor arguments.")
    if to_checksum_address(monitoring_service.functions.token().call()) != token_address:
        raise RuntimeError("MonitoringService has a wrong token address onchain.")
    if token_address != constructor_arguments[0]:
        raise RuntimeError("MonitoringService received a wrong token address during construction")

    if (
        to_checksum_address(monitoring_service.functions.service_registry().call())
        != service_registry_address
    ):
        raise RuntimeError("MonitoringService has a wrong ServiceRegistry address onchain.")
    if service_registry_address != constructor_arguments[1]:
        raise RuntimeError("MonitoringService received a wrong address during construction.")
    if (
        to_checksum_address(monitoring_service.functions.user_deposit().call())
        != user_deposit_address
    ):
        raise RuntimeError("MonitoringService has a wrong UserDeposit address onchain.")
    if user_deposit_address != constructor_arguments[2]:
        raise RuntimeError(
            "MonitoringService received a wrong UserDeposit address during construction."
        )
    if (
        to_checksum_address(monitoring_service.functions.token_network_registry().call())
        != token_network_registry_address
    ):
        raise RuntimeError("MonitoringService has a wrong TokenNetworkRegistry address onchain.")
    if token_network_registry_address != constructor_arguments[3]:
        raise RuntimeError(
            "MonitoringService received a wrong TokenNetworkRegistry address during construction."
        )


def _verify_one_to_n_deployment(
    one_to_n: Contract,
    constructor_arguments: List,
    user_deposit_address: HexAddress,
    service_registry_address: HexAddress,
    chain_id: int,
) -> None:
    """Check an onchain deployment of OneToN and constructor arguments"""
    if to_checksum_address(one_to_n.functions.deposit_contract().call()) != user_deposit_address:
        raise RuntimeError("OneToN has a wrong UserDeposit address onchain.")
    if user_deposit_address != constructor_arguments[0]:
        raise RuntimeError("OneToN received a wrong UserDeposit address during construction.")
    if chain_id != constructor_arguments[1]:
        raise RuntimeError("OneToN received a wrong chain ID during construction.")
    if service_registry_address != constructor_arguments[2]:
        raise RuntimeError("OneToN received a wrong ServiceRegistry address during construction.")
    if len(constructor_arguments) != 3:
        raise RuntimeError("OneToN received a wrong number of constructor arguments.")


def _verify_service_registry_deployment(
    service_registry: Contract, constructor_arguments: List, token_address: HexAddress
) -> None:
    """Check an onchain deployment of ServiceRegistry and constructor arguments"""
    if len(constructor_arguments) != 8:
        raise RuntimeError(
            "ServiceRegistry was deployed with a wrong number of constructor arguments"
        )
    if to_checksum_address(service_registry.functions.token().call()) != token_address:
        raise RuntimeError("ServiceRegistry has a wrong token address")
    if token_address != constructor_arguments[0]:
        raise RuntimeError(
            f"expected token address {token_address} "
            f"but the constructor argument for {CONTRACT_SERVICE_REGISTRY} is "
            f"{constructor_arguments[0]}"
        )
    controller_onchain = to_checksum_address(service_registry.functions.controller().call())
    if controller_onchain != constructor_arguments[1]:
        raise RuntimeError(
            f"the deployment data contains the controller address {constructor_arguments[1]} "
            f"but the contract remembers {controller_onchain} onchain."
        )
    # All other parameters can change after the deployment, so the checks are omitted.



from enum import IntEnum
from typing import NamedTuple

from eth_typing import HexAddress, HexStr
from eth_utils import decode_hex
from eth_utils.units import units

from raiden_contracts.utils.signature import private_key_to_address
from raiden_contracts.utils.type_aliases import (
    AdditionalHash,
    BalanceHash,
    Locksroot,
    PrivateKey,
    Signature,
)

UINT256_MAX = 2 ** 256 - 1
NOT_ADDRESS = "0xaaa"
FAKE_ADDRESS = HexAddress(HexStr("0x00112233445566778899aabbccddeeff00112233"))
EMPTY_HEXADDRESS = "0x0000000000000000000000000000000000000000"
EMPTY_BALANCE_HASH = BalanceHash(b"\x00" * 32)
EMPTY_ADDITIONAL_HASH = AdditionalHash(b"\x00" * 32)
EMPTY_LOCKSROOT = Locksroot(b"\x00" * 32)
EMPTY_SIGNATURE = Signature(b"\x00" * 65)
passphrase = "0"
FAUCET_PRIVATE_KEY = PrivateKey(
    decode_hex("0xaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa")
)
FAUCET_ADDRESS = private_key_to_address(FAUCET_PRIVATE_KEY)
FAUCET_ALLOWANCE = 100 * int(units["ether"])
DEPLOYER_ADDRESS = FAUCET_ADDRESS
NONEXISTENT_LOCKSROOT = b"\x00" * 32
SECONDS_PER_DAY = 60 * 60 * 24

# Constants for ServiceRegistry testing
SERVICE_DEPOSIT = 5000 * (10 ** 18)
DEFAULT_BUMP_NUMERATOR = 6
DEFAULT_BUMP_DENOMINATOR = 5
DEFAULT_DECAY_CONSTANT = 200 * SECONDS_PER_DAY
DEFAULT_REGISTRATION_DURATION = 180 * SECONDS_PER_DAY
DEFAULT_MIN_PRICE = 1000


class LockIndex(IntEnum):
    EXPIRATION = 0
    AMOUNT = 1
    SECRETHASH = 2
    SECRET = 3


class OnchainBalanceProof(NamedTuple):
    balance_hash: bytes
    nonce: int
    additional_hash: bytes
    original_signature: bytes



from logging import getLogger
from typing import Optional, Tuple

import click
from eth_typing import URI, HexAddress, HexStr
from eth_utils import encode_hex
from web3.contract import Contract

from raiden_contracts.constants import (
    CONTRACT_CUSTOM_TOKEN,
    CONTRACT_TOKEN_NETWORK,
    CONTRACT_TOKEN_NETWORK_REGISTRY,
    DEPLOY_SETTLE_TIMEOUT_MAX,
    DEPLOY_SETTLE_TIMEOUT_MIN,
    MAX_ETH_CHANNEL_PARTICIPANT,
    MAX_ETH_TOKEN_NETWORK,
)
from raiden_contracts.deploy.__main__ import setup_ctx
from raiden_contracts.deploy.contract_deployer import ContractDeployer
from raiden_contracts.tests.utils import call_and_transact
from raiden_contracts.utils.transaction import check_successful_tx

log = getLogger(__name__)


@click.command()
@click.option("--private-key", required=True, help="Path to a private key store.")
@click.option(
    "--rpc-provider", default="http://127.0.0.1:8545", help="Address of the Ethereum RPC provider"
)
@click.option("--wait", default=300, help="Max tx wait time in s.")
@click.option("--gas-price", default=5, type=int, help="Gas price to use in gwei")
@click.option("--gas-limit", default=5_500_000)
@click.pass_context
def deprecation_test(
    ctx: click.Context,
    private_key: str,
    rpc_provider: URI,
    wait: int,
    gas_price: int,
    gas_limit: int,
) -> None:
    """Turn on the deprecation switch and see channel opening fails"""
    setup_ctx(
        ctx=ctx,
        private_key=private_key,
        password_file=None,
        rpc_provider=rpc_provider,
        wait=wait,
        gas_price=gas_price,
        gas_limit=gas_limit,
    )
    deployer = ctx.obj["deployer"]

    # We deploy the Raiden Network contracts and register a token network
    token_amount = MAX_ETH_CHANNEL_PARTICIPANT * 6
    channel_participant_deposit_limit = MAX_ETH_CHANNEL_PARTICIPANT
    token_network_deposit_limit = MAX_ETH_TOKEN_NETWORK
    (_, token_network, _) = deprecation_test_setup(
        deployer=deployer,
        token_amount=token_amount,
        channel_participant_deposit_limit=channel_participant_deposit_limit,
        token_network_deposit_limit=token_network_deposit_limit,
    )

    log.info("Checking that channels can be opened and deposits can be made.")

    # Check that we can open channels and deposit on behalf of A and B
    # Some arbitrary Ethereum addresses
    A = HexAddress(HexStr("0x6AA63296FA94975017244769F00F0c64DB7d7115"))
    B = HexAddress(HexStr("0xc9a4fad99B6d7D3e48D18d2585470cd8f27FA61e"))
    channel_identifier = open_and_deposit(A=A, B=B, token_network=token_network, deployer=deployer)
    log.info("Seding transaction to activate the deprecation switch.")

    # Activate deprecation switch
    assert token_network.functions.safety_deprecation_switch().call() is False
    txhash = deployer.transact(token_network.functions.deprecate())
    log.debug(f"Deprecation txHash={encode_hex(txhash)}")
    assert token_network.functions.safety_deprecation_switch().call() is True

    log.info("Checking that channels cannot be opened anymore and no more deposits are allowed.")

    # Check that we cannot open more channels or deposit
    C = HexAddress(HexStr("0x5a23cedB607684118ccf7906dF3e24Efd2964719"))
    D = HexAddress(HexStr("0x3827B9cDc68f061aa614F1b97E23664ef3b9220A"))
    open_and_deposit(
        A=C,
        B=D,
        token_network=token_network,
        deployer=deployer,
        channel_identifier=channel_identifier,
        txn_success_status=False,
    )

    log.info("Deprecation switch test OK.")


def deprecation_test_setup(
    deployer: ContractDeployer,
    token_amount: int,
    channel_participant_deposit_limit: int,
    token_network_deposit_limit: int,
) -> Tuple:
    deployed_contracts = deployer.deploy_raiden_contracts(
        max_num_of_token_networks=1,
        reuse_secret_registry_from_deploy_file=None,
        settle_timeout_min=DEPLOY_SETTLE_TIMEOUT_MIN,
        settle_timeout_max=DEPLOY_SETTLE_TIMEOUT_MAX,
    )["contracts"]

    token_network_registry_abi = deployer.contract_manager.get_contract_abi(
        CONTRACT_TOKEN_NETWORK_REGISTRY
    )
    token_network_registry = deployer.web3.eth.contract(
        abi=token_network_registry_abi,
        address=deployed_contracts[CONTRACT_TOKEN_NETWORK_REGISTRY]["address"],
    )

    token_decimals = 18
    multiplier = 10 ** token_decimals
    token_supply = 10 ** 6 * multiplier
    token_amount = int(token_amount * multiplier)

    deployed_token = deployer.deploy_token_contract(
        token_supply, token_decimals, "TestToken", "TTT", CONTRACT_CUSTOM_TOKEN
    )
    token_address = deployed_token[CONTRACT_CUSTOM_TOKEN]
    token_abi = deployer.contract_manager.get_contract_abi(CONTRACT_CUSTOM_TOKEN)
    token_contract = deployer.web3.eth.contract(abi=token_abi, address=token_address)

    # Mint some tokens for the owner
    txhash = call_and_transact(token_contract.functions.mint(token_amount), deployer.transaction)

    log.debug(f"Minting tokens txHash={encode_hex(txhash)}")
    check_successful_tx(deployer.web3, txhash, deployer.wait)
    assert token_contract.functions.balanceOf(deployer.owner).call() >= token_amount

    abi = deployer.contract_manager.get_contract_abi(CONTRACT_TOKEN_NETWORK_REGISTRY)
    token_network_address = deployer.register_token_network(
        token_registry_abi=abi,
        token_registry_address=deployed_contracts[CONTRACT_TOKEN_NETWORK_REGISTRY]["address"],
        token_address=token_address,
        channel_participant_deposit_limit=channel_participant_deposit_limit,
        token_network_deposit_limit=token_network_deposit_limit,
    )["token_network_address"]

    token_network_abi = deployer.contract_manager.get_contract_abi(CONTRACT_TOKEN_NETWORK)
    token_network = deployer.web3.eth.contract(
        abi=token_network_abi, address=token_network_address
    )
    log.info(
        f"Registered the token and created a TokenNetwork contract at {token_network_address}."
    )

    txhash = call_and_transact(
        token_contract.functions.approve(token_network.address, token_amount),
        deployer.transaction,
    )
    log.debug(f"Approving tokens for the TokenNetwork contract txHash={encode_hex(txhash)}")
    check_successful_tx(deployer.web3, txhash, deployer.wait)

    assert (
        token_contract.functions.allowance(deployer.owner, token_network.address).call()
        >= token_amount
    )
    log.info(
        f"Approved {token_amount} tokens for the TokenNetwork contract "
        f"from owner {deployer.owner}."
    )

    return (token_network_registry, token_network, token_contract)


def open_and_deposit(
    A: HexAddress,
    B: HexAddress,
    token_network: Contract,
    deployer: ContractDeployer,
    channel_identifier: Optional[int] = None,
    txn_success_status: bool = True,
) -> int:
    try:
        txhash = call_and_transact(
            token_network.functions.openChannel(
                participant1=A, participant2=B, settle_timeout=DEPLOY_SETTLE_TIMEOUT_MIN
            ),
            deployer.transaction,
        )
        log.debug(f"Opening a channel between {A} and {B} txHash={encode_hex(txhash)}")
        check_successful_tx(web3=deployer.web3, txid=txhash, timeout=deployer.wait)

        # Get the channel identifier
        channel_identifier = token_network.functions.getChannelIdentifier(A, B).call()
        success_status = True
    except ValueError as ex:
        success_status = False
        log.info(f"Cannot open a new channel {ex}")

    assert (
        txn_success_status == success_status
    ), f"openChannel txn status is {success_status} instead of {txn_success_status}"

    assert channel_identifier is not None
    try:
        txhash = call_and_transact(
            token_network.functions.setTotalDeposit(
                channel_identifier=channel_identifier,
                participant=A,
                total_deposit=int(MAX_ETH_CHANNEL_PARTICIPANT / 2),
                partner=B,
            ),
            deployer.transaction,
        )
        log.debug(
            f"Depositing {MAX_ETH_CHANNEL_PARTICIPANT} tokens for {A} in a channel with "
            f"identifier={channel_identifier} and partner= {B} txHash={encode_hex(txhash)}"
        )
        check_successful_tx(web3=deployer.web3, txid=txhash, timeout=deployer.wait)
        success_status = True
    except ValueError as ex:
        success_status = False
        log.info(f"Cannot deposit more tokens in channel={channel_identifier}, {ex}")

    assert (
        txn_success_status == success_status
    ), f"setTotalDeposit txn status is {success_status} instead of {txn_success_status}"

    try:
        txhash = call_and_transact(
            token_network.functions.setTotalDeposit(
                channel_identifier=channel_identifier,
                participant=B,
                total_deposit=int(MAX_ETH_CHANNEL_PARTICIPANT / 2),
                partner=A,
            ),
            deployer.transaction,
        )
        log.debug(
            f"Depositing {MAX_ETH_CHANNEL_PARTICIPANT} tokens for {B} in a channel with "
            f"identifier={channel_identifier} and partner= {A} txHash={encode_hex(txhash)}"
        )
        check_successful_tx(web3=deployer.web3, txid=txhash, timeout=deployer.wait)
        success_status = True
    except ValueError as ex:
        success_status = False
        log.info(f"Cannot deposit more tokens in channel={channel_identifier}, {ex}")

    assert (
        txn_success_status == success_status
    ), f"setTotalDeposit txn status is {success_status} instead of {txn_success_status}"

    return channel_identifier


if __name__ == "__main__":
    # pylint: disable=E1120
    deprecation_test()



import getpass
import json
import logging
import os
import stat
from pathlib import Path
from typing import Optional

from eth_keyfile import decode_keyfile_json
from eth_utils import decode_hex, is_hex

from raiden_contracts.utils.type_aliases import PrivateKey

log = logging.getLogger(__name__)

# This file was copied from
# https://github.com/raiden-network/raiden-libs/blob/9902dcdb74b8d18a232df3f1e1dc5442882419fe/raiden_libs/utils/private_key.py
# during the deprecation of raiden-lib.


def check_permission_safety(path: Path) -> bool:
    """Check if the file at the given path is safe to use as a state file.

    This checks that group and others have no permissions on the file and that the current user is
    the owner.
    """
    f_stats = os.stat(path)
    return (f_stats.st_mode & (stat.S_IRWXG | stat.S_IRWXO)) == 0 and f_stats.st_uid == os.getuid()


def get_private_key(key_path: Path, password_path: Optional[Path] = None) -> Optional[PrivateKey]:
    """Open a JSON-encoded private key and return it

    If a password file is provided, uses it to decrypt the key. If not, the
    password is asked interactively. Raw hex-encoded private keys are supported,
    but deprecated."""

    if not key_path:
        log.critical(f"key_path has to be something but got {key_path}")
        return None

    if not os.path.exists(key_path):
        log.critical("%s: no such file", key_path)
        return None

    if not check_permission_safety(key_path):
        log.critical("Private key file %s must be readable only by its owner.", key_path)
        return None

    if password_path and not check_permission_safety(password_path):
        log.critical("Password file %s must be readable only by its owner.", password_path)
        return None

    with open(key_path) as keyfile:
        raw_keyfile = keyfile.readline().strip()

        if is_hex(raw_keyfile) and len(decode_hex(raw_keyfile)) == 32:
            log.warning("Private key in raw format. Consider switching to JSON-encoded")
            return PrivateKey(decode_hex(raw_keyfile))
        else:
            keyfile.seek(0)
            try:
                json_data = json.load(keyfile)
                if password_path:
                    with open(password_path) as password_file:
                        password = password_file.readline().strip()
                else:
                    password = getpass.getpass("Enter the private key password: ")
                if json_data["crypto"]["kdf"] == "pbkdf2":
                    password = password.encode()  # type: ignore
                return PrivateKey(decode_keyfile_json(json_data, password))
            except ValueError:
                log.critical("Invalid private key format or password!")
                return None



from functools import reduce
from hashlib import sha256
from os import urandom
from random import randint
from typing import Collection, Iterable, List, NamedTuple, Optional, Tuple

from eth_abi import encode_abi
from eth_utils import keccak
from web3 import Web3

from raiden_contracts.constants import TEST_SETTLE_TIMEOUT_MIN

PendingTransfersTree = NamedTuple(
    "PendingTransfersTree",
    [
        ("transfers", List[List]),
        ("unlockable", List[List]),
        ("expired", List[List]),
        ("packed_transfers", bytes),
        ("hash_of_packed_transfers", bytes),
        ("locked_amount", int),
    ],
)


def get_pending_transfers_tree_with_generated_lists(
    web3: Web3,
    unlockable_amount: int,
    expired_amount: int,
    min_expiration_delta: Optional[int] = None,
    max_expiration_delta: Optional[int] = None,
) -> PendingTransfersTree:
    return get_pending_transfers_tree(
        web3=web3,
        unlockable_amounts=get_random_values_for_sum(unlockable_amount),
        expired_amounts=get_random_values_for_sum(expired_amount),
        min_expiration_delta=min_expiration_delta,
        max_expiration_delta=max_expiration_delta,
    )


def get_pending_transfers_tree(
    web3: Web3,
    unlockable_amounts: Collection[int],
    expired_amounts: Collection[int],
    min_expiration_delta: Optional[int] = None,
    max_expiration_delta: Optional[int] = None,
) -> PendingTransfersTree:
    types = ["uint256", "uint256", "bytes32"]
    packed_transfers = b""
    (unlockable_locks, expired_locks) = get_pending_transfers(
        web3=web3,
        unlockable_amounts=unlockable_amounts,
        expired_amounts=expired_amounts,
        min_expiration_delta=min_expiration_delta,
        max_expiration_delta=max_expiration_delta,
    )

    pending_transfers = unlockable_locks + expired_locks

    hashed_pending_transfers = [
        Web3.solidityKeccak(types, transfer_data[:-1])  # pylint: disable=E1120
        for transfer_data in pending_transfers
    ]

    if len(pending_transfers) > 0:
        hashed_pending_transfers, pending_transfers = zip(
            *sorted(zip(hashed_pending_transfers, pending_transfers))
        )
        pending_transfers = list(pending_transfers)
        packed_transfers = get_packed_transfers(pending_transfers=pending_transfers, types=types)

    locked_amount = get_locked_amount(pending_transfers)

    return PendingTransfersTree(
        transfers=pending_transfers,
        unlockable=unlockable_locks,
        expired=expired_locks,
        packed_transfers=packed_transfers,
        hash_of_packed_transfers=keccak(packed_transfers),
        locked_amount=locked_amount,
    )


def get_pending_transfers(
    web3: Web3,
    unlockable_amounts: Collection[int],
    expired_amounts: Iterable[int],
    min_expiration_delta: Optional[int],
    max_expiration_delta: Optional[int],
) -> Tuple:
    current_block = web3.eth.block_number
    if expired_amounts is None:
        expired_amounts = []
    min_expiration_delta = min_expiration_delta or (len(unlockable_amounts) + 1)
    max_expiration_delta = max_expiration_delta or (min_expiration_delta + TEST_SETTLE_TIMEOUT_MIN)
    unlockable_locks = [
        [
            current_block + randint(min_expiration_delta, max_expiration_delta),
            amount,
            *random_secret(),
        ]
        for amount in unlockable_amounts
    ]
    expired_locks = [[current_block, amount, *random_secret()] for amount in expired_amounts]
    return (unlockable_locks, expired_locks)


def get_packed_transfers(pending_transfers: Collection, types: List) -> bytes:
    packed_transfers: List[bytes] = [encode_abi(types, x[:-1]) for x in pending_transfers]
    return reduce((lambda x, y: x + y), packed_transfers)


def get_locked_amount(pending_transfers: List) -> int:
    return reduce((lambda x, y: x + y[1]), pending_transfers, 0)


def random_secret() -> Tuple:
    secret = urandom(32)
    hasher = sha256(secret)
    return (hasher.digest(), secret)


def get_random_values_for_sum(values_sum: int) -> List[int]:
    amount = 0
    values = []
    while amount < values_sum:
        value = randint(1, values_sum - amount)
        values.append(value)
        amount += value
    return values
